import itertools
import warnings
from abc import abstractmethod, ABC
from collections import defaultdict

from negmas import (
    Contract,
    Breach,
    Negotiator,
    MechanismState,
    AgentMechanismInterface,
    RenegotiationRequest,
    JavaSAONegotiator,
    UtilityFunction,
    JavaUtilityFunction,
    _ShadowAgentMechanismInterface,
)
from .common import DEFAULT_NEGOTIATOR, ProductionReport
from negmas.apps.scml.simulators import FactorySimulator, FastFactorySimulator
from negmas.apps.scml.simulators import storage_as_array, temporary_transaction
from negmas.common import NamedObject
from negmas.events import Notification
from negmas.helpers import get_class, instantiate
from negmas.helpers import snake_case
from negmas.java import (
    JavaCallerMixin,
    to_java,
    from_java,
    to_dict,
    java_link,
    PYTHON_CLASS_IDENTIFIER,
)
from negmas.outcomes import Issue, Outcome
from negmas.sao import _ShadowSAONegotiator
from negmas.utilities import UtilityValue
from .agent import SCMLAgent
from .awi import _ShadowSCMLAWI, SCMLAWI
from .common import (
    SCMLAgreement,
    Factory,
    CFP,
    Loan,
    ProductionFailure,
    FinancialReport,
)
from .consumers import ScheduleDrivenConsumer, ConsumptionProfile
from .schedulers import Scheduler, ScheduleInfo, GreedyScheduler

if True:
    from typing import (
        Dict,
        Iterable,
        Any,
        Callable,
        Collection,
        Type,
        List,
        Optional,
        Union,
    )

__all__ = [
    "FactoryManager",
    "DoNothingFactoryManager",
    "GreedyFactoryManager",
    "JavaFactoryManager",
    "JavaDoNothingFactoryManager",
    "JavaGreedyFactoryManager",
    "JavaDummyMiddleMan",
]


class FactoryManager(SCMLAgent, ABC):
    """Base factory manager class that will be inherited by participant negmas in ANAC 2019.

    The agent can access the world simulation in one of two ways:

    1. Attributes and methods available in the Agent-World-Interface (See `SCMLAWI` documentation for those).
    2. Attributes and methods in the `FactoryManager` object itself. All factory managers will have the following
       attributes and methods that simplify the interaction with the world simulation. Some of these attributes/methods
       are convenient ways to access functionality already available in the agent's internal `SCMLAWI`.

    **Attributes**

    *Agent information*

    - `id` : The unique ID assigned to this agent. This is unique system-wide and is what is used in contracts, CFPs,
      etc.
    - `name`: A name of the agent used for display purposes only. The simulator never accesses or uses this name except
      in printing and logging.
    - `uuid` : Another name of the `id` .
    - `type_name` : A string giving the type of the agent (as a fully qualified python class name).

    *Capabilities/Profiles*

    - `line_profiles` : A mapping specifying for each line index, all the profiles that can be run on it
    - `process_profiles` : A mapping specifying for each `Process` index, all the profiles used to run it in the factory
    - `producing` : Mapping from a product index to all manufacturing processes that can generate it
    - `consuming` : Mapping from a product index to all manufacturing processes that can consume it
    - `compiled_profiles` : All the profiles to be used by the factory belonging to this agent compiled to use process
      indices
    - `max_storage` : Maximum storage available to the agent. Zero, None or float('inf') all indicate unlimited storage.

    *Production Graph* (also accessible through *awi*)

    - `products` : List of products in the system
    - `processes` : List of processes in the system

    *Helper Objects*

    - `awi` : The `SCMLAWI` instance assigned to this agent. It can be used to interact with the simulation (See
      `SCMLAWI` documentation).
    - `simulator` : A `FactorySimulator` object that can be used to simulate what happens in the `Factory` assigned to
      this agent when given operations are conducted (e.g. production, paying money, etc).

    *Negotiations/Contracts*

    - `requested_negotiations` : A dynamic list of negotiations currently requested by the agent but not started.
      *Correct management of this list is only possible if the agent **always** uses `request_negotiation` method of
      this class (see methods later) rather than directly calling request_method on the `SCMLAWI` ( `awi` ) member.
    - `running_negotiations` : A dynamic list of negotiations currently running involving this agent.
      *Correct management of this list is only possible if the agent **always** uses `request_negotiation` method of
      this class (see methods later) rather than directly calling request_method on the `SCMLAWI` ( `awi` ) member.
    - `unsigned_contracts` : A dynamic list of negotiations contracts concluded involving this agent but not yet signed.
      *Correct management of this list is only possible if the agent **always** uses `request_negotiation` method of
      this class (see methods later) rather than directly calling request_method on the `SCMLAWI` ( `awi` ) member.

    *Simulation attributes* (also accessible through *awi*)

    - `transportation_delay` : The transportation delay in the system.
    - `current_step` : Current simulation step.
    - `immediate_negotiations` : Whether or not negotiations start immediately upon registration (default is to start on
      the next production step)
    - `negotiation_speed_multiple` : The number of negotiation rounds (steps) conducted in a single production step
    - `transportation_delay` : Transportation delay in the system. Default is zero


    **Methods** (Callable by the agent)

    *Actions on the world*

    - `request_negotiation` : Called to request a negotiation based on a `CFP` .

    *Scheduling and simulation helpers*

    - `can_expect_agreement` : Checks if it is possible in principle to get an agreement on this CFP by the time it
      becomes executable.


    **Callbacks** (Callable by the simulation)

    *Decision callbacks* (Called to make decisions)

        - Negotiation and Contracts

            - `respond_to_negotiation_request` : Decide whether or not to engage in a negotiation on a `CFP` that was
              published earlier by this factory manager. If accepted, the agent should return a `SAONegotiator` object.
            - `sign_contract` : Decide whether or not to sign the contract. If accepted, the agent should return its own ID.
            - `confirm_contract_execution` : Decide whether or not to go on with executing a contract that the agent already
              signed. If rejected (by returning `False` ), a refusal-to-execute breach will be recorded.

        - Breach related

            - `confirm_partial_execution` : Decide whether the agent agrees to partial execution. Called only when the
              the partner of this agent commits a partial breach (of level < 1) and this agent commits no breaches.
            - `set_renegotiation_agenda` : Decide what are the issues and ranges of acceptable values to re-negotiate about.
              Called only in case of breaches.
            - `respond_to_renegotiation_request` : Decide whether or not to engage in a re-negotiation.

        - Financial

            - `confirm_loan` : Decide whether or not to accept an offered loan. *In ANAC 2019 league, loans are not allowed
              and this callback will never be called by the simulator.

    *Time-dependent callbacks* (Information callback called at predefined times)

        - `init` : Called once before any production or negotiations to initiate the agent.
        - `step` : Called at every production step.

    *Information callbacks* (Called to inform the agent about events)

        - CFP related

            - `on_new_cfp` : Called whenever a `CFP` on a `Product` for which the agent has already registered interest
              (using `register_interest` method of its `awi`) is published. By default all agents register interest in the
              products they can consume or produce according to their profiles.
            - `on_remove_cfp` : Called whenever a `CFP` on a `Product` for which the agent has already registered interest
              (using `register_interest` method of its `awi`) is removed from the bulletin-board.

        - Negotiation related

            - `on_neg_request_accepted` : Called when a negotiation request of the agent is accepted
            - `on_neg_request_rejected` : Called when a negotiation request of the agent is rejected
            - `on_negotiation_success` : Called when a negotiation of which the agent is a party succeeds with an agreement.
            - `on_negotiation_failure` : Called when a negotiation of which the agent is a party ends without agreement.

        - Contract related

            - `on_contract_cancelled` : Called whenever a `Contract` of which the agent is a party is cancelled because the
              other party refused to sign it.
            - `on_contract_signed` : Called whenever a `Contract` of which the agent is a party is signed by both patners.
            - `on_contract_nullified` : Called whenever a `Contract` of which the agent is a party is nullified by the
              simulator as a part of bankruptcy processing.
            - `on_contract_executed` : Called when a contract executes completely and successfully.
            - `on_contract_breached` : Called when a contract is breached after complete contract processing.

        - Production and factory related

            - `on_production_failure` : Called whenever a scheduled production (see `SCMLAWI` for production commands)
              fails
            - `on_inventory_change` : Called whenever there is a change in the inventory (something is moved in or out
              or out of storage due to an event other than production (e.g. contract execution).
            - `on_cash_transfer` : Called whenever cash is transferred to or from the factory's wallet.

        - About other agents

            - `on_agent_bankrupt` : Called whenever another agent goes bankrupt
            - `on_new_report` : Called whenever a new report of another agent for which this agent has registered interest
              is published. Interest is registered using the agent's `awi` 's `receive_financial_reports` method.

    """

    def __init__(
        self,
        name=None,
        simulator_type: Union[str, Type[FactorySimulator]] = FastFactorySimulator,
    ):
        super().__init__(name=name)
        self.transportation_delay = 0
        """Transportation delay in the world"""
        self.simulator: Optional[FactorySimulator] = None
        """The simulator used by this agent"""
        self.simulator_type: Type[FactorySimulator] = get_class(
            simulator_type, scope=globals()
        )
        """Simulator type (as a class)"""
        self.current_step = 0
        """Current simulation step"""
        self.max_storage: int = 0
        """Maximum storage available to the agent"""

    def init_(self):
        state: Factory = self.awi.state
        self.current_step = state.next_step
        self.max_storage = state.max_storage
        self.simulator = self.simulator_type(
            initial_wallet=state.wallet,
            initial_storage=state.storage,
            n_steps=self.awi.n_steps,
            n_products=len(self.awi.products),
            profiles=state.profiles,
            max_storage=self.max_storage,
        )
        super().init_()

    def step_(self):
        state = self.awi.state
        self.simulator.set_state(
            self.current_step,
            wallet=state.wallet,
            loans=state.loans,
            storage=storage_as_array(state.storage, n_products=len(self.products)),
            line_schedules=state.line_schedules,
        )
        self.current_step += 1
        self.step()

    @abstractmethod
    def on_production_failure(self, failures: List[ProductionFailure]) -> None:
        """Called with a list of `ProductionFailure` records on production failure."""

    @abstractmethod
    def on_production_success(self, reports: List[ProductionReport]) -> None:
        """Called with a list of `ProductionReport` records on production success"""

    class Java:
        implements = ["jnegmas.apps.scml.factory_managers.FactoryManager"]


class DoNothingFactoryManager(FactoryManager):
    """The default factory manager that will be implemented by the committee of ANAC-SCML 2019"""

    def init(self):
        pass

    def step(self):
        pass

    def on_neg_request_rejected(self, req_id: str, by: Optional[List[str]]):
        pass

    def on_neg_request_accepted(self, req_id: str, mechanism: AgentMechanismInterface):
        pass

    def on_negotiation_failure(
        self,
        partners: List[str],
        annotation: Dict[str, Any],
        mechanism: AgentMechanismInterface,
        state: MechanismState,
    ) -> None:
        pass

    def on_negotiation_success(
        self, contract: Contract, mechanism: AgentMechanismInterface
    ) -> None:
        pass

    def on_contract_signed(self, contract: Contract) -> None:
        pass

    def on_contract_cancelled(self, contract: Contract, rejectors: List[str]) -> None:
        pass

    def on_contract_executed(self, contract: Contract) -> None:
        pass

    def on_contract_breached(
        self, contract: Contract, breaches: List[Breach], resolution: Optional[Contract]
    ) -> None:
        pass

    def sign_contract(self, contract: Contract) -> Optional[str]:
        return self.id

    def on_contract_nullified(
        self, contract: Contract, bankrupt_partner: str, compensation: float
    ) -> None:
        pass

    def on_agent_bankrupt(self, agent_id: str) -> None:
        pass

    def confirm_partial_execution(
        self, contract: Contract, breaches: List[Breach]
    ) -> bool:
        return True

    def on_remove_cfp(self, cfp: "CFP") -> None:
        pass

    def on_production_failure(self, failures: List[ProductionFailure]) -> None:
        pass

    def respond_to_negotiation_request(
        self, cfp: "CFP", partner: str
    ) -> Optional[Negotiator]:
        return None

    def confirm_contract_execution(self, contract: Contract) -> bool:
        return True

    def set_renegotiation_agenda(
        self, contract: Contract, breaches: List[Breach]
    ) -> Optional[RenegotiationRequest]:
        return None

    def respond_to_renegotiation_request(
        self, contract: Contract, breaches: List[Breach], agenda: RenegotiationRequest
    ) -> Optional[Negotiator]:
        return None

    def confirm_loan(self, loan: Loan, bankrupt_if_rejected: bool) -> bool:
        """called by the world manager to confirm a loan if needed by the buyer of a contract that is about to be
        breached"""
        return bankrupt_if_rejected

    def on_new_cfp(self, cfp: "CFP") -> None:
        pass

    def on_inventory_change(self, product: int, quantity: int, cause: str) -> None:
        pass

    def on_production_success(self, reports: List[ProductionReport]) -> None:
        pass

    def on_cash_transfer(self, amount: float, cause: str) -> None:
        pass

    def on_new_report(self, report: FinancialReport):
        pass


class GreedyFactoryManager(DoNothingFactoryManager):
    """The default factory manager that will be implemented by the committee of ANAC-SCML 2019"""

    def on_production_failure(self, failures: List[ProductionFailure]) -> None:
        pass

    def on_production_success(self, reports: List[ProductionReport]) -> None:
        pass

    def confirm_loan(self, loan: Loan, bankrupt_if_rejected: bool) -> bool:
        return bankrupt_if_rejected

    def confirm_contract_execution(self, contract: Contract) -> bool:
        return True

    def set_renegotiation_agenda(
        self, contract: Contract, breaches: List[Breach]
    ) -> Optional[RenegotiationRequest]:
        return None

    def respond_to_renegotiation_request(
        self, contract: Contract, breaches: List[Breach], agenda: RenegotiationRequest
    ) -> Optional[Negotiator]:
        return None

    def __init__(
        self,
        name=None,
        simulator_type: Union[str, Type[FactorySimulator]] = FastFactorySimulator,
        scheduler_type: Union[str, Type[Scheduler]] = GreedyScheduler,
        scheduler_params: Optional[Dict[str, Any]] = None,
        optimism: float = 0.0,
        negotiator_type: Union[str, Type[Negotiator]] = DEFAULT_NEGOTIATOR,
        negotiator_params: Optional[Dict[str, Any]] = None,
        n_retrials=5,
        use_consumer=True,
        reactive=True,
        sign_only_guaranteed_contracts=False,
        riskiness=0.0,
        max_insurance_premium: float = 0.1,
        reserved_value: float = None,
    ):
        super().__init__(name=name, simulator_type=simulator_type)
        self.negotiator_type = get_class(negotiator_type, scope=globals())
        self.negotiator_params = (
            negotiator_params if negotiator_params is not None else {}
        )
        self.optimism = optimism
        self.ufun_factory: Union[
            Type[NegotiatorUtility], Callable[[Any, Any], NegotiatorUtility]
        ]
        if optimism < 1e-6:
            self.ufun_factory = PessimisticNegotiatorUtility
        elif optimism > 1 - 1e-6:
            self.ufun_factory = OptimisticNegotiatorUtility
        else:
            self.ufun_factory: NegotiatorUtility = lambda agent, annotation: AveragingNegotiatorUtility(
                agent=agent, annotation=annotation, optimism=self.optimism
            )
        if max_insurance_premium < 0.0:
            warnings.warn(
                f"Negative max insurance ({max_insurance_premium}) is deprecated. Set max_insurance_premium = inf "
                f"for always buying and max_insurance_premium = 0.0 for never buying. Will continue assuming inf"
            )
            max_insurance_premium = float("inf")
        self.__reserved_value = reserved_value
        self.max_insurance_premium = max_insurance_premium
        self.n_retrials = n_retrials
        self.n_neg_trials: Dict[str, int] = defaultdict(int)
        self.consumer = None
        self.use_consumer = use_consumer
        self.reactive = reactive
        self.sign_only_guaranteed_contracts = sign_only_guaranteed_contracts
        self.contract_schedules: Dict[str, ScheduleInfo] = {}
        self.riskiness = riskiness
        self.negotiation_margin = int(round(n_retrials * max(0.0, 1.0 - riskiness)))
        self.scheduler_type: Type[Scheduler] = get_class(
            scheduler_type, scope=globals()
        )
        self.scheduler: Scheduler = None
        self.scheduler_params: Dict[
            str, Any
        ] = scheduler_params if scheduler_params is not None else {}

    def total_utility(self, contracts: Collection[Contract] = ()) -> float:
        """Calculates the total utility for the agent of a collection of contracts"""
        if self.scheduler is None:
            raise ValueError("Cannot calculate total utility without a scheduler")
        min_concluded_at = self.awi.current_step
        min_sign_at = min_concluded_at + self.awi.default_signing_delay
        with temporary_transaction(self.scheduler):
            schedule = self.scheduler.schedule(
                contracts=contracts,
                assume_no_further_negotiations=False,
                ensure_storage_for=self.transportation_delay,
                start_at=min_sign_at,
            )
        if not schedule.valid:
            return float("-inf")
        return schedule.final_balance

    def init(self):
        self.negotiation_margin = max(
            self.negotiation_margin,
            int(round(len(self.products) * max(0.0, 1.0 - self.riskiness))),
        )
        if self.use_consumer:
            # @todo add the parameters of the consumption profile as parameters of the greedy factory manager
            profiles = dict(
                zip(
                    self.consuming.keys(),
                    (
                        ConsumptionProfile(schedule=[_] * self.awi.n_steps)
                        for _ in itertools.repeat(0)
                    ),
                )
            )
            self.consumer: ScheduleDrivenConsumer = ScheduleDrivenConsumer(
                profiles=profiles,
                consumption_horizon=self.awi.n_steps,
                immediate_cfp_update=True,
                name=self.name,
            )
            self.consumer.id = self.id
            self.consumer.awi = self.awi
            self.consumer.init_()
        self.scheduler = self.scheduler_type(
            manager_id=self.id,
            awi=self.awi,
            max_insurance_premium=self.max_insurance_premium,
            **self.scheduler_params,
        )
        self.scheduler.init(
            simulator=self.simulator,
            products=self.products,
            processes=self.processes,
            producing=self.producing,
            profiles=self.compiled_profiles,
        )

    def respond_to_negotiation_request(
        self, cfp: "CFP", partner: str
    ) -> Optional[Negotiator]:
        if self.awi.is_bankrupt(partner):
            return None
        if self.use_consumer:
            return self.consumer.respond_to_negotiation_request(
                cfp=cfp, partner=partner
            )
        else:
            ufun_ = self.ufun_factory(
                self, self._create_annotation(cfp=cfp, partner=partner)
            )
            ufun_.reserved_value = self.__reserved_value
            neg = self.negotiator_type(
                name=self.name + "*" + partner[:4], **self.negotiator_params, ufun=ufun_
            )
            return neg

    def on_negotiation_success(
        self, contract: Contract, mechanism: AgentMechanismInterface
    ):
        if self.use_consumer:
            self.consumer.on_negotiation_success(contract, mechanism)

    def on_negotiation_failure(
        self,
        partners: List[str],
        annotation: Dict[str, Any],
        mechanism: AgentMechanismInterface,
        state: MechanismState,
    ) -> None:
        if self.use_consumer:
            self.consumer.on_negotiation_failure(partners, annotation, mechanism, state)
        cfp = annotation["cfp"]
        thiscfp = self.awi.bb_query(section="cfps", query=cfp.id, query_keys=True)
        if (
            cfp.publisher != self.id
            and thiscfp is not None
            and len(thiscfp) > 0
            and self.n_neg_trials[cfp.id] < self.n_retrials
        ):
            self.awi.logdebug(f"Renegotiating {self.n_neg_trials[cfp.id]} on {cfp}")
            self.n_neg_trials[cfp.id] += 1
            self.on_new_cfp(cfp=annotation["cfp"])

    def _execute_schedule(self, schedule: ScheduleInfo, contract: Contract) -> None:
        if self.simulator is None:
            raise ValueError("No factory simulator is defined")
        awi: SCMLAWI = self.awi
        total = contract.agreement["unit_price"] * contract.agreement["quantity"]
        product = contract.annotation["cfp"].product
        if contract.annotation["buyer"] == self.id:
            self.simulator.buy(
                product=product,
                quantity=contract.agreement["quantity"],
                price=total,
                t=contract.agreement["time"],
            )
            if total <= 0 or self.max_insurance_premium <= 0.0 or contract is None:
                return
            relative_premium = awi.evaluate_insurance(contract=contract)
            if relative_premium is None:
                return
            premium = relative_premium * total
            if relative_premium <= self.max_insurance_premium:
                self.awi.logdebug(
                    f"{self.name} buys insurance @ {premium:0.02} ({relative_premium:0.02%}) for {str(contract)}"
                )
                awi.buy_insurance(contract=contract)
                self.simulator.pay(premium, self.awi.current_step)
            return
        # I am a seller
        self.simulator.sell(
            product=product,
            quantity=contract.agreement["quantity"],
            price=total,
            t=contract.agreement["time"],
        )
        for job in schedule.jobs:
            if job.action == "run":
                awi.schedule_job(job, contract=contract)
            elif job.action == "stop":
                awi.stop_production(
                    line=job.line,
                    step=job.time,
                    contract=contract,
                    override=job.override,
                )
            else:
                awi.schedule_job(job, contract=contract)
            self.simulator.schedule(job=job, override=False)
        for need in schedule.needs:
            if need.quantity_to_buy <= 0:
                continue
            product_id = need.product
            # self.simulator.reserve(product=product_id, quantity=need.quantity_to_buy, t=need.step)
            if self.use_consumer:
                self.consumer.profiles[product_id].schedule[
                    need.step
                ] += need.quantity_to_buy
                self.consumer.register_product_cfps(
                    p=product_id,
                    t=need.step,
                    profile=self.consumer.profiles[product_id],
                )
                continue
            product = self.products[product_id]
            if product.catalog_price is None:
                price_range = (0.0, 100.0)
            else:
                price_range = (0.5, 1.5 * product.catalog_price)
            # @todo check this. This error is raised sometimes
            if need.step < awi.current_step:
                continue
                # raise ValueError(f'need {need} at {need.step} while running at step {awi.current_step}')
            time = (
                need.step
                if self.max_storage is not None
                else (awi.current_step, need.step)
            )
            cfp = CFP(
                is_buy=True,
                publisher=self.id,
                product=product_id,
                time=time,
                unit_price=price_range,
                quantity=(1, int(1.1 * need.quantity_to_buy)),
            )
            awi.register_cfp(cfp)

    def sign_contract(self, contract: Contract):
        if any(self.awi.is_bankrupt(partner) for partner in contract.partners):
            return None
        signature = self.id
        with temporary_transaction(self.scheduler):
            schedule = self.scheduler.schedule(
                assume_no_further_negotiations=False,
                contracts=[contract],
                ensure_storage_for=self.transportation_delay,
                start_at=self.awi.current_step + 1,
            )

        if self.sign_only_guaranteed_contracts and (
            not schedule.valid or len(schedule.needs) > 1
        ):
            self.awi.logdebug(
                f"{self.name} refused to sign contract {contract.id} because it cannot be scheduled"
            )
            return None
        # if schedule.final_balance <= self.simulator.final_balance:
        #     self.awi.logdebug(f'{self.name} refused to sign contract {contract.id} because it is not expected '
        #                       f'to lead to profit')
        #     return None
        if schedule.valid:
            profit = schedule.final_balance - self.simulator.final_balance
            self.awi.logdebug(
                f"{self.name} singing contract {contract.id} expecting "
                f'{-profit if profit < 0 else profit} {"loss" if profit < 0 else "profit"}'
            )
        else:
            self.awi.logdebug(
                f"{self.name} singing contract {contract.id} expecting breach"
            )
            return None

        self.contract_schedules[contract.id] = schedule
        return signature

    def on_contract_signed(self, contract: Contract):
        if contract.annotation["buyer"] == self.id and self.use_consumer:
            self.consumer.on_contract_signed(contract)
        schedule = self.contract_schedules[contract.id]
        if schedule is not None and schedule.valid:
            self._execute_schedule(schedule=schedule, contract=contract)
        if contract.annotation["buyer"] != self.id or not self.use_consumer:
            for negotiation in self._running_negotiations.values():
                self.notify(
                    negotiation.negotiator,
                    Notification(type="ufun_modified", data=None),
                )

    def _process_buy_cfp(self, cfp: "CFP") -> None:
        if cfp.publisher == self.id:
            return
        if self.awi.is_bankrupt(cfp.publisher):
            return
        if self.simulator is None or not self.can_expect_agreement(
            cfp=cfp, margin=self.negotiation_margin
        ):
            return
        if not self.can_produce(cfp=cfp):
            return
        neg = self.negotiator_type(
            name=self.name + ">" + cfp.publisher[:4], **self.negotiator_params
        )
        ufun = self.ufun_factory(self, self._create_annotation(cfp=cfp))
        ufun.reserved_value = self.__reserved_value
        self.request_negotiation(negotiator=neg, cfp=cfp, ufun=ufun)
        # normalize(, outcomes=cfp.outcomes, infeasible_cutoff=-1)

    def _process_sell_cfp(self, cfp: "CFP"):
        if self.awi.is_bankrupt(cfp.publisher):
            return None
        if self.use_consumer:
            self.consumer.on_new_cfp(cfp=cfp)

    def on_new_cfp(self, cfp: "CFP") -> None:
        if not self.reactive:
            return
        if cfp.satisfies(
            query={"is_buy": True, "products": list(self.producing.keys())}
        ):
            self._process_buy_cfp(cfp)
        if cfp.satisfies(
            query={"is_buy": False, "products": list(self.consuming.keys())}
        ):
            self._process_sell_cfp(cfp)

    def step(self):
        if self.use_consumer:
            self.consumer.step()
        if self.reactive:
            return
        # 0. remove all my CFPs
        # self.awi.bb_remove(section='cfps', query={'publisher': self})

        # respond to interesting CFPs
        # todo: should check time and sort products by interest etc
        cfps = self.awi.bb_query(
            section="cfps", query={"products": self.producing.keys(), "is_buy": True}
        )
        if cfps is None:
            return
        for cfp in cfps.values():
            self._process_buy_cfp(cfp)

    def can_produce(self, cfp: CFP, assume_no_further_negotiations=False) -> bool:
        """Whether or not we can produce the required item in time"""
        if cfp.product not in self.producing.keys():
            return False
        agreement = SCMLAgreement(
            time=cfp.max_time, unit_price=cfp.max_unit_price, quantity=cfp.min_quantity
        )
        min_concluded_at = self.awi.current_step + 1 - int(self.immediate_negotiations)
        min_sign_at = min_concluded_at + self.awi.default_signing_delay
        if cfp.max_time < min_sign_at + 1:  # 1 is minimum time to produce the product
            return False
        with temporary_transaction(self.scheduler):
            schedule = self.scheduler.schedule(
                contracts=[
                    Contract(
                        partners=[self.id, cfp.publisher],
                        agreement=agreement,
                        annotation=self._create_annotation(cfp=cfp),
                        issues=cfp.issues,
                        signed_at=min_sign_at,
                        concluded_at=min_concluded_at,
                    )
                ],
                ensure_storage_for=self.transportation_delay,
                assume_no_further_negotiations=assume_no_further_negotiations,
                start_at=min_sign_at,
            )
        return schedule.valid and self.can_secure_needs(
            schedule=schedule, step=self.awi.current_step
        )

    def can_secure_needs(self, schedule: ScheduleInfo, step: int):
        """
        Finds if it is possible in principle to arrange these needs at the given time.

        Args:
            schedule:
            step:

        Returns:

        """
        needs = schedule.needs
        if len(needs) < 1:
            return True
        for need in needs:
            if need.quantity_to_buy > 0 and need.step < step + 1 - int(
                self.immediate_negotiations
            ):  # @todo check this
                return False
        return True


TotalUtilityFun = Callable[[Collection[Contract]], float]


class NegotiatorUtility(UtilityFunction):
    """The utility function of a negotiator."""

    def __init__(
        self,
        agent: GreedyFactoryManager,
        annotation: Dict[str, Any],
        name: Optional[str] = None,
        avoid_free_sales: bool = True,
    ):
        if name is None:
            name = (
                agent.name
                + "*"
                + "*".join(_ for _ in annotation["partners"] if _ != agent.id)
            )
        super().__init__(name=name)
        self.agent = agent
        self.annotation = annotation
        self.avoid_free_sales = avoid_free_sales

    def _contracts(self, agreements: Iterable[SCMLAgreement]) -> Collection[Contract]:
        """Converts agreements/outcomes into contracts"""
        if self.ami is None:
            raise ValueError("No annotation is stored (No mechanism info)")
        annotation = self.ami.annotation
        return [
            Contract(
                partners=annotation["partners"],
                agreement=a,
                annotation=annotation,
                issues=self.ami.issues,
            )
            for a in agreements
        ]

    def _contract(self, agreement: SCMLAgreement) -> Contract:
        """Converts an agreement/outcome into a contract"""
        annotation = self.annotation
        return Contract(
            partners=annotation["partners"],
            agreement=agreement,
            annotation=annotation,
            issues=annotation["cfp"].issues,
        )

    def _free_sale(self, agreement: SCMLAgreement) -> bool:
        return (
            (
                self.annotation["seller"] == self.agent.id
                and agreement["unit_price"] < 1e-6
            )
            if self.avoid_free_sales
            else False
        )

    def __call__(self, outcome: Outcome) -> Optional[UtilityValue]:
        if outcome is None:
            return float("-inf")
        if isinstance(outcome, dict):
            return self.call(agreement=SCMLAgreement(**outcome))
        if isinstance(outcome, SCMLAgreement):
            return self.call(agreement=outcome)
        raise ValueError(f"Outcome: {outcome} cannot be converted to an SCMLAgreement")

    @abstractmethod
    def call(self, agreement: SCMLAgreement) -> Optional[UtilityValue]:
        """Called to evaluate a agreement"""

    def xml(self, issues: List[Issue]) -> str:
        return "NegotiatorUtility has not xml representation"


class PessimisticNegotiatorUtility(NegotiatorUtility):
    """The utility function of a negotiator that assumes other negotiations currently open will fail."""

    def call(self, agreement: SCMLAgreement) -> Optional[UtilityValue]:
        """An offer will be a tuple of one value which in turn will be a list of contracts"""
        if self._free_sale(agreement):
            return float("-inf")
        # contracts = self.agent.contracts
        # hypothetical = list(contracts)
        # hypothetical.append(self._contract(agreement))
        hypothetical = [self._contract(agreement)]
        base_util = self.agent.simulator.final_balance
        hypothetical = self.agent.total_utility(hypothetical)
        if hypothetical < 0:
            return float("-inf")
        return hypothetical - base_util


class OptimisticNegotiatorUtility(NegotiatorUtility):
    """The utility function of a negotiator that assumes other negotiations currently open will succeed."""

    def call(self, agreement: SCMLAgreement) -> Optional[UtilityValue]:
        if self._free_sale(agreement):
            return float("-inf")
        # contracts = self.agent.contracts
        # hypothetical = list(contracts)
        # hypothetical.append(self._contract(agreement))
        hypothetical = [self._contract(agreement)]
        for negotiation in self.agent.running_negotiations:  # type: ignore
            negotiator = negotiation.negotiator
            current_offer = negotiator.my_last_proposal
            if current_offer is not None:
                hypothetical.append(self._contract(current_offer))
        base_util = self.agent.simulator.final_balance
        hypothetical = self.agent.total_utility(list(hypothetical))
        if hypothetical < 0:
            return float("-inf")
        return hypothetical - base_util


class AveragingNegotiatorUtility(NegotiatorUtility):
    """A utility function that combines optimistic and pessimistic evaluators linearly using adjustable weight"""

    def __init__(
        self,
        agent: GreedyFactoryManager,
        annotation: Dict[str, Any],
        name: Optional[str] = None,
        optimism: float = 0.5,
        avoid_free_sale: bool = True,
    ):
        NamedObject.__init__(self=self, name=name)
        self.avoid_free_sales = avoid_free_sale
        self.optimism = optimism
        self.optimistic = OptimisticNegotiatorUtility(
            agent=agent, annotation=annotation, avoid_free_sales=avoid_free_sale
        )
        self.pessimistic = PessimisticNegotiatorUtility(
            agent=agent, annotation=annotation, avoid_free_sales=avoid_free_sale
        )

    def call(self, agreement: SCMLAgreement) -> Optional[UtilityValue]:
        if self._free_sale(agreement):
            return float("-inf")
        opt, pess = self.optimistic(agreement), self.pessimistic(agreement)
        if opt is None or pess is None:
            return None
        return self.optimism * opt + (1 - self.optimism) * pess


class JavaFactoryManager(FactoryManager, JavaCallerMixin):
    """Allows factory managers implemented in Java (using jnegmas) to participate in SCML worlds.

    Objects of this class is used to represent a java object to the python environment. This means that they *MUST* have
    the same interface as a python class (first class in the inheritance list). The `JavaCallerMixin` is used to enable
    it to connect to the java object it is representing.

    """

    def on_production_success(self, reports: List[ProductionReport]) -> None:
        self._java_object.onProductionSuccess(to_java(reports))

    def on_contract_executed(self, contract: Contract) -> None:
        self._java_object.onContractExecuted(to_java(contract))

    def on_contract_breached(
        self, contract: Contract, breaches: List[Breach], resolution: Optional[Contract]
    ) -> None:
        self._java_object.onContractBreached(
            to_java(contract), to_java(breaches), to_java(resolution)
        )

    def on_inventory_change(self, product: int, quantity: int, cause: str) -> None:
        self._java_object.onInventoryChange(product, quantity, cause)

    def on_cash_transfer(self, amount: float, cause: str) -> None:
        self._java_object.onCashTransfer(amount, cause)

    @property
    def type_name(self):
        """Overrides type name to give the internal java type name"""
        return "j" + snake_case(
            self._java_class_name.replace(
                "jnegmas.apps.scml.factory_managers.", ""
            ).replace("FactoryManager", "")
        )

    @property
    def awi(self):
        return self._awi

    @awi.setter
    def awi(self, value):
        self._awi = value
        if self.python_shadow is not self:
            self.python_shadow._awi = value
        self.java_awi = _ShadowSCMLAWI(value)
        self._java_object.setAWI(self.java_awi)

    def init(self):
        if self.python_shadow is not self:
            self.python_shadow.simulator = self.simulator
        self._java_object.setSimulator(self.simulator)
        self._java_object.init()

    def step(self):
        return self._java_object.init()

    def on_neg_request_rejected(self, req_id: str, by: Optional[List[str]]):
        return self._java_object.onNegRequestRejected(req_id, by)

    def on_neg_request_accepted(self, req_id: str, mechanism: AgentMechanismInterface):
        return self._java_object.onNegRequestAccepted(
            req_id, java_link(_ShadowAgentMechanismInterface(mechanism))
        )

    def on_new_cfp(self, cfp: "CFP"):
        return from_java(self._java_object.onNewCFP(to_java(cfp)))

    def on_remove_cfp(self, cfp: "CFP"):
        return self._java_object.onRemoveCFP(to_java(cfp))

    def on_contract_nullified(
        self, contract: Contract, bankrupt_partner: str, compensation: float
    ) -> None:
        self._java_object.onContractNullified(
            to_java(contract), bankrupt_partner, compensation
        )

    def on_agent_bankrupt(self, agent_id: str) -> None:
        self._java_object.onAgentBankrupt(agent_id)

    def confirm_partial_execution(
        self, contract: Contract, breaches: List[Breach]
    ) -> bool:
        return self._java_object.confirmParialExecution(
            to_java(contract), to_java(breaches)
        )

    def on_production_failure(self, failures: List[ProductionFailure]) -> None:
        return self._java_object.onProductionFailure(to_java(failures))

    def confirm_loan(self, loan: Loan, bankrupt_if_rejected: bool) -> bool:
        return self._java_object.confirmLoan(to_java(loan), bankrupt_if_rejected)

    def confirm_contract_execution(self, contract: Contract) -> bool:
        return self._java_object.confirmContractExecution(to_java(contract))

    def respond_to_negotiation_request(
        self, cfp: "CFP", partner: str
    ) -> Optional[Negotiator]:
        result = self._java_object.respondToNegotiationRequest(to_java(cfp), partner)
        if result is None:
            return result
        return JavaSAONegotiator(java_object=result, java_class_name=None)

    def on_negotiation_failure(
        self,
        partners: List[str],
        annotation: Dict[str, Any],
        mechanism: AgentMechanismInterface,
        state: MechanismState,
    ) -> None:
        return self._java_object.onNegotiationFailure(
            to_java(partners),
            annotation,
            java_link(_ShadowAgentMechanismInterface(mechanism)),
            to_java(state),
        )

    def on_negotiation_success(
        self, contract: Contract, mechanism: AgentMechanismInterface
    ) -> None:
        return self._java_object.onNegotiationSuccess(
            to_java(contract), java_link(_ShadowAgentMechanismInterface(mechanism))
        )

    def on_contract_signed(self, contract: Contract) -> None:
        return self._java_object.onContractSigned(to_java(contract))

    def on_contract_cancelled(self, contract: Contract, rejectors: List[str]) -> None:
        return self._java_object.onContractCancelled(
            to_java(contract), to_java(rejectors)
        )

    def on_new_report(self, report: FinancialReport):
        pass

    def sign_contract(self, contract: Contract) -> Optional[str]:
        return from_java(self._java_object.signContract(to_java(contract)))

    def set_renegotiation_agenda(
        self, contract: Contract, breaches: List[Breach]
    ) -> Optional[RenegotiationRequest]:
        return from_java(
            self._java_object.setRenegotiationAgenda(
                to_java(contract), to_java(breaches)
            )
        )

    def respond_to_renegotiation_request(
        self, contract: Contract, breaches: List[Breach], agenda: RenegotiationRequest
    ) -> Optional[Negotiator]:
        return from_java(
            self._java_object.respondToRenegotiationRequest(
                to_java(contract), to_java(breaches), to_java(agenda)
            )
        )

    # handy constructors
    @classmethod
    def do_nothing_manager(cls):
        return JavaFactoryManager(
            java_class_name="jnegmas.apps.scml.factory_managers.DoNothingFactoryManager"
        )

    @classmethod
    def greedy_manager(cls):
        return JavaFactoryManager(
            java_class_name="jnegmas.apps.scml.factory_managers.GreedyFactoryManager"
        )

    def __init__(
        self,
        java_object=None,
        java_class_name: str = None,
        python_shadow: Optional[FactoryManager] = None,
        auto_load_java: bool = False,
        name=None,
        simulator_type: Union[str, Type[FactorySimulator]] = FastFactorySimulator,
    ):
        super().__init__(name=name, simulator_type=simulator_type)
        self.java_awi = None
        if java_class_name is not None:
            stem = java_class_name.split(".")[-1]
            if stem.endswith("GreedyFactoryManager") or stem.endswith("GFM"):
                python_shadow = GreedyFactoryManager(
                    name=self.name, simulator_type=self.simulator_type
                )
                python_shadow.id = self.id
        if python_shadow is None:
            python_shadow = self
        else:
            python_shadow = python_shadow
        self.python_shadow = python_shadow
        self._callback_shadow = self.python_shadow
        self.init_java_bridge(
            java_object=java_object,
            java_class_name=java_class_name,
            auto_load_java=auto_load_java,
            python_shadow_object=self,
        )
        if java_object is None:
            map = to_dict(self)
            map.pop(PYTHON_CLASS_IDENTIFIER, None)
            map["simulatorType"] = self.simulator_type.__class__.__name__
            self._java_object.fromMap(to_java(map))

    def getNegotiationRequests(self):
        return to_java(self.requested_negotiations)

    def getRunningNegotiations(self):
        return to_java(self.running_negotiations)

    def requestNegotiation(
        self, cfp: CFP, negotiator: Negotiator = None, ufun: UtilityFunction = None
    ) -> bool:
        return self.request_negotiation(
            from_java(cfp),
            JavaSAONegotiator(negotiator, None),
            JavaUtilityFunction(ufun, None),
        )

    def getLineProfiles(self):
        return to_java(self.line_profiles)

    def getProducing(self):
        return to_java(self.producing)

    def getConsuming(self):
        return to_java(self.consuming)

    def getCompiledProfiles(self):
        return to_java(self.compiled_profiles)

    def getProducts(self):
        return to_java(self.products)

    def getProcesses(self):
        return to_java(self.processes)

    def getContracts(self):
        return to_java(self.contracts)

    def getRequestedNegotiations(self):
        return to_java(self.requested_negotiations)

    def getRunningNegotiations(self):
        return to_java(self.running_negotiations)

    def requestNegotiation(
        self, cfp: CFP, negotiator: Negotiator = None, ufun: UtilityFunction = None
    ) -> bool:
        return self.request_negotiation(
            from_java(cfp),
            JavaSAONegotiator(negotiator, None),
            JavaUtilityFunction(ufun, None),
        )

    def getID(self):
        return self.id

    def setID(self, value):
        self.id = value

    def getName(self):
        return self.name

    def setName(self, value):
        self.name = value

    def initPython(self):
        return self._callback_shadow.init()

    def stepPython(self):
        return self._callback_shadow.step()

    def onNegRequestRejected(self, req_id, rejectors):
        return self._callback_shadow.on_neg_request_rejected(
            req_id, from_java(rejectors)
        )

    def onNegRequestAccepted(self, req_id, mechanism):
        return self._callback_shadow.on_neg_request_accepted(
            req_id, from_java(mechanism)
        )

    def onNewCFP(self, cfp):
        return self._callback_shadow.on_new_cfp(from_java(cfp))

    def onRemoveCFP(self, cfp):
        return self._callback_shadow.on_remove_cfp(from_java(cfp))

    def onContractNullified(self, contract, bankruptPartner, compensation):
        return self._callback_shadow.on_contract_nullified(
            from_java(contract), bankruptPartner, compensation
        )

    def onAgentBankrupt(self, agentId):
        return self._callback_shadow.on_agent_bankrupt(agentId)

    def confirmPartialExecution(self, contract, breaches):
        return self._callback_shadow.confirm_partial_execution(
            from_java(contract), from_java(breaches)
        )

    def onProductionFailure(self, failures):
        return self._callback_shadow.on_production_failure(from_java(failures))

    def onProductionSuccess(self, reports) -> None:
        self._callback_shadow.on_production_success(from_java(reports))

    def onContractExecuted(self, contract: Contract) -> None:
        self._callback_shadow.on_contract_executed(from_java(contract))

    def onContractBreached(
        self, contract: Contract, breaches: List[Breach], resolution: Optional[Contract]
    ) -> None:
        self._callback_shadow.on_contract_breached(
            from_java(contract), from_java(breaches), from_java(resolution)
        )

    def onInventoryChange(self, product: int, quantity: int, cause: str) -> None:
        self._callback_shadow.on_inventory_change(product, quantity, cause)

    def onCashTransfer(self, amount: float, cause: str) -> None:
        self._callback_shadow.on_cash_transfer(amount, cause)

    def confirmLoan(self, loan, bankruptIfRejected):
        return self._callback_shadow.confirm_loan(from_java(loan), bankruptIfRejected)

    def confirmContractExecution(self, contract):
        return self._callback_shadow.confirm_contract_execution(from_java(contract))

    def respondToNegotiationRequest(self, cfp, partner):
        result = self._callback_shadow.respond_to_negotiation_request(
            from_java(cfp), partner
        )
        if result is None:
            return None
        return _ShadowSAONegotiator(result)

    def onNegotiationFailure(self, partners, annotation, mechanism, state):
        return self._callback_shadow.on_negotiation_failure(
            from_java(partners),
            from_java(annotation),
            from_java(mechanism),
            from_java(state),
        )

    def onNegotiationSuccess(self, contract, mechanism):
        return self._callback_shadow.on_negotiation_success(
            from_java(contract), from_java(mechanism)
        )

    def onContractSigned(self, contract):
        return self._callback_shadow.on_contract_signed(from_java(contract))

    def onContractCancelled(self, contract, rejectors):
        return self._callback_shadow.on_contract_cancelled(
            from_java(contract), from_java(rejectors)
        )

    def onNewReport(self, report):
        return self._callback_shadow.on_new_report(from_java(report))

    def signContract(self, contract):
        return self._callback_shadow.sign_contract(from_java(contract))

    def setRenegotiationAgenda(self, contract, breaches):
        return to_java(
            self._callback_shadow.set_renegotiation_agenda(
                from_java(contract), from_java(breaches)
            )
        )

    def respondToRenegotiationRequest(self, contract, breaches, agenda):
        result = self._callback_shadow.respond_to_renegotiation_request(
            from_java(contract), from_java(breaches), from_java(agenda)
        )
        if result is None:
            return result
        return _ShadowSAONegotiator(result)

    class Java:
        implements = ["jnegmas.apps.scml.factory_managers.FactoryManager"]


class JavaDoNothingFactoryManager(JavaFactoryManager):
    def __init__(
        self,
        auto_load_java: bool = False,
        name=None,
        simulator_type: Union[str, Type[FactorySimulator]] = FastFactorySimulator,
    ):
        super().__init__(
            name=name,
            simulator_type=simulator_type,
            auto_load_java=auto_load_java,
            java_class_name="jnegmas.apps.scml.factory_managers.DoNothingFactoryManager",
        )


class JavaGreedyFactoryManager(JavaFactoryManager):
    def __init__(
        self,
        auto_load_java: bool = False,
        name=None,
        simulator_type: Union[str, Type[FactorySimulator]] = FastFactorySimulator,
    ):
        super().__init__(
            name=name,
            simulator_type=simulator_type,
            auto_load_java=auto_load_java,
            java_class_name="jnegmas.apps.scml.factory_managers.GreedyFactoryManager",
            python_shadow=lambda: GreedyFactoryManager(
                name=name, simulator_type=self.simulator_type
            ),
        )


class JavaDummyMiddleMan(JavaFactoryManager):
    def __init__(
        self,
        auto_load_java: bool = False,
        name=None,
        simulator_type: Union[str, Type[FactorySimulator]] = FastFactorySimulator,
    ):
        super().__init__(
            name=name,
            simulator_type=simulator_type,
            auto_load_java=auto_load_java,
            java_class_name="jnegmas.apps.scml.factory_managers.DummyMiddleMan",
        )



"""Defines import/export functionality
"""
import functools
import operator
import os
import shutil
from typing import List, Optional, Tuple, Union, Dict, Callable, Iterable

import numpy as np
import pkg_resources

from negmas.generics import ivalues

__all__ = [
    "load_genius_domain",
    "load_genius_domain_from_folder",
    "convert_genius_domain_from_folder",
    "convert_genius_domain",
    "find_domain_and_utility_files",
    "get_domain_issues",
]

import xml.etree.ElementTree as ET
from os import listdir
from negmas import Issue, enumerate_outcomes, make_discounted_ufun
from negmas import Negotiator
from negmas import UtilityFunction
from negmas import SAOMechanism, AspirationNegotiator


def get_domain_issues(
    domain_file_name: str,
    force_single_issue=False,
    max_n_outcomes: int = 1e6,
    n_discretization: Optional[int] = None,
    keep_issue_names=True,
    keep_value_names=True,
    safe_parsing=False,
) -> Union[Dict[str, Issue], List[Issue]]:
    """
    Returns the issues of a given XML domain (Genius Format)

    Args:
        domain_file_name:
        force_single_issue:
        max_n_outcomes:
        n_discretization:
        keep_issue_names:
        keep_value_names:
        safe_parsing:

    Returns:
        List or Dict of issues

    """
    issues, issues_details, mechanism = None, None, None
    if domain_file_name is not None:
        domain_file_name = str(domain_file_name)
        issues_details, _ = Issue.from_genius(
            domain_file_name,
            force_single_issue=False,
            keep_issue_names=True,
            keep_value_names=True,
            safe_parsing=safe_parsing,
            n_discretization=n_discretization,
        )
        if force_single_issue:
            issues, _ = Issue.from_genius(
                domain_file_name,
                force_single_issue=force_single_issue,
                keep_issue_names=keep_issue_names,
                keep_value_names=keep_value_names,
                max_n_outcomes=max_n_outcomes,
                n_discretization=n_discretization,
            )
            if issues is None:
                return []
        else:
            issues, _ = Issue.from_genius(
                domain_file_name,
                force_single_issue=force_single_issue,
                keep_issue_names=keep_issue_names,
                keep_value_names=keep_value_names,
                safe_parsing=safe_parsing,
                n_discretization=n_discretization,
            )
    return issues if not force_single_issue else [issues]


def load_genius_domain(
    domain_file_name: str,
    utility_file_names: Optional[List[str]] = None,
    agent_factories: Optional[
        Union[Callable[[], Negotiator], List[Callable[[], Negotiator]]]
    ] = None,
    force_single_issue=False,
    force_numeric=False,
    cache_and_discretize_outcomes=False,
    max_n_outcomes: int = 1e6,
    n_discretization: Optional[int] = None,
    keep_issue_names=True,
    keep_value_names=True,
    normalize_utilities=True,
    n_steps=None,
    time_limit=3 * 60,  # GENIUS uses 3min time limit by default
    max_n_agents=None,
    dynamic_entry=True,
    safe_parsing=False,
    ignore_reserved=False,
    ignore_discount=False,
) -> Tuple[Optional[SAOMechanism], List[dict], Union[Dict[str, Issue], List[Issue]]]:
    """
    Loads a genius domain, creates appropriate negotiators if necessary

    Args:
        domain_file_name:
        utility_file_names:
        agent_factories:
        force_single_issue:
        cache_and_discretize_outcomes:
        max_n_outcomes:
        n_discretization:
        keep_issue_names:
        keep_value_names:
        normalize_utilities:
        n_steps:
        time_limit:
        max_n_agents:
        dynamic_entry:
        safe_parsing:
        ignore_reserved:
        ignore_discount:

    Returns:
        - mechanism (SAOMechanism): A mechanism for the given issues
        - agent_info (List[Dict]): All Negotiator functions from the given file
        - issues Union[Issue, Dict[str, Issue], List[Issue]]] : The issues

    """
    issues, issues_details, mechanism = None, None, None
    if domain_file_name is not None:
        domain_file_name = str(domain_file_name)
        issues_details, _ = Issue.from_genius(
            domain_file_name,
            force_single_issue=False,
            keep_issue_names=True,
            keep_value_names=True,
            safe_parsing=safe_parsing,
            n_discretization=n_discretization,
        )
        if force_single_issue:
            issues, _ = Issue.from_genius(
                domain_file_name,
                force_single_issue=force_single_issue,
                keep_issue_names=keep_issue_names,
                keep_value_names=keep_value_names,
                max_n_outcomes=max_n_outcomes,
                n_discretization=n_discretization,
                force_numeric=force_numeric,
            )
            if issues is None:
                return None, [], []
        else:
            issues, _ = Issue.from_genius(
                domain_file_name,
                force_single_issue=force_single_issue,
                keep_issue_names=keep_issue_names,
                keep_value_names=keep_value_names,
                safe_parsing=safe_parsing,
                n_discretization=n_discretization,
                force_numeric=force_numeric,
            )

    agent_info = []
    if utility_file_names is None:
        utility_file_names = []
    utility_file_names = [str(_) for _ in utility_file_names]
    for ufname in utility_file_names:
        utility, discount_factor = UtilityFunction.from_genius(
            file_name=ufname,
            force_single_issue=force_single_issue,
            keep_issue_names=keep_issue_names,
            keep_value_names=keep_value_names,
            normalize_utility=normalize_utilities,
            domain_issues=issues_details,
            safe_parsing=safe_parsing,
            max_n_outcomes=max_n_outcomes,
            ignore_discount=ignore_discount,
            ignore_reserved=ignore_reserved,
            force_numeric=force_numeric,
        )
        agent_info.append(
            {
                "ufun": utility,
                "ufun_name": ufname,
                "reserved_value_func": utility.reserved_value
                if utility is not None
                else 0.0,
                "discount_factor": discount_factor,
            }
        )
    outcomes = None
    try:
        if force_single_issue or cache_and_discretize_outcomes or len(issues) == 1:
            n_outcomes: float = functools.reduce(
                operator.mul,
                (
                    float(_.cardinality())
                    if not _.is_continuous()
                    else float(n_discretization)
                    if n_discretization is not None
                    else np.inf
                    for _ in ivalues(issues)
                ),
                1.0,
            )
            if n_outcomes < max_n_outcomes:
                outcomes = enumerate_outcomes(issues, keep_issue_names=keep_issue_names)
    except ValueError:
        pass
    if domain_file_name is not None:
        mechanism_name = (
            domain_file_name.split("/")[-1][:-4]
            .replace("-domain", "")
            .replace("_domain", "")
            .replace("domain", "")
        )
        mechanism = SAOMechanism(
            issues=issues,
            outcomes=outcomes,
            n_steps=n_steps,
            time_limit=time_limit,
            max_n_agents=max_n_agents,
            dynamic_entry=dynamic_entry,
            name=mechanism_name,
            keep_issue_names=keep_issue_names,
        )
        if agent_info is not None and len(agent_info) > 0:
            for info in agent_info:
                info["ufun"] = (
                    info["ufun"]
                    if info["discount_factor"] is None or info["discount_factor"] == 1.0
                    else make_discounted_ufun(
                        ufun=info["ufun"],
                        ami=mechanism.ami,
                        discount_per_round=info["discount_factor"],
                        power_per_round=1.0,
                    )
                )

    if agent_factories is not None and agent_info is not None and len(agent_info) > 0:
        if not isinstance(agent_factories, Iterable):
            agent_factories = [agent_factories] * len(agent_info)
        agents = [factory() for factory in agent_factories[0 : len(agent_info)]]
        for a, info in zip(agents, agent_info):
            mechanism.add(a, ufun=info["ufun"])

    return mechanism, agent_info, (issues if not force_single_issue else [issues])


def load_genius_domain_from_folder(
    folder_name: str,
    agent_factories: Optional[
        Union[Callable[[], Negotiator], List[Callable[[], Negotiator]]]
    ] = None,
    force_single_issue=False,
    force_numeric=False,
    cache_and_discretize_outcomes=False,
    max_n_outcomes: int = 1e6,
    n_discretization: Optional[int] = None,
    keep_issue_names=True,
    keep_value_names=True,
    normalize_utilities=True,
    n_steps=None,
    time_limit=60,  # GENIUS uses 3min time limit by default
    max_n_agents=None,
    dynamic_entry=True,
    safe_parsing=False,
    ignore_reserved=False,
    ignore_discount=False,
) -> Tuple[Optional[SAOMechanism], List[dict], Union[Dict[str, Issue], List[Issue]]]:
    """
    Loads a genius domain from a folder. See ``load_genius_domain`` for more details.

    Args:
        folder_name:
        agent_factories:
        force_single_issue:
        cache_and_discretize_outcomes:
        max_n_outcomes:
        n_discretization:
        keep_issue_names:
        keep_value_names:
        normalize_utilities:
        n_steps:
        time_limit:
        max_n_agents:
        dynamic_entry:
        safe_parsing:
        ignore_reserved:
        ignore_discount:

    Returns:
        - mechanism (SAOMechanism): A mechanism for the given issues
        - agent_info (List[Dict]): All Negotiator functions from the given file
        - issues Union[Issue, Dict[str, Issue], List[Issue]]] : The issues

    Examples:

        >>> folder_name = pkg_resources.resource_filename('negmas', resource_name='tests/data/10issues')
        >>> mechanism, negotiators, issues = load_genius_domain_from_folder(folder_name
        ...                             , force_single_issue=False, keep_issue_names=False
        ...                             , keep_value_names=False, normalize_utilities=False)
        >>> print(len(issues))
        10
        >>> print(len(negotiators))
        2
        >>> print([type(a['ufun']) for a in negotiators])
        [<class 'negmas.utilities.HyperRectangleUtilityFunction'>, <class 'negmas.utilities.HyperRectangleUtilityFunction'>]
        >>> print(negotiators[0]['ufun'].outcome_ranges[0])
        {1: (7.0, 9.0), 3: (2.0, 7.0), 5: (0.0, 8.0), 8: (0.0, 7.0)}
        >>> print(negotiators[0]['ufun'].mappings[0])
        97.0
        >>> u = negotiators[0]['ufun']
        >>> print(u(tuple([0.0] * len(issues))))
        0.0
        >>> print(u(tuple([0.5] * len(issues))))
        186.0

        Try loading and running a domain with predetermined agents:
        >>> mechanism, agents, issues = load_genius_domain_from_folder(
        ...                             pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop')
        ...                             , agent_factories=AspirationNegotiator
        ...                             , force_single_issue=True, keep_issue_names=False
        ...                             , keep_value_names=False)
        >>> state = mechanism.run()
        >>> state.agreement is not None
        True


        >>> mechanism, negotiators, issues = load_genius_domain_from_folder(
        ...                             pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop'))

        >>> len(issues), len(negotiators)
        (3, 2)

        >>> [type(a['ufun']) for a in negotiators]
        [<class 'negmas.utilities.LinearUtilityAggregationFunction'>, <class 'negmas.utilities.LinearUtilityAggregationFunction'>]

        >>> mechanism, negotiators, issues = load_genius_domain_from_folder(
        ...                             pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop')
        ...                             , force_single_issue=True, keep_issue_names=False
        ...                             , keep_value_names=False)
        >>> len(issues), len(negotiators)
        (1, 2)
        >>> [type(a['ufun']) for a in negotiators]
        [<class 'negmas.utilities.MappingUtilityFunction'>, <class 'negmas.utilities.MappingUtilityFunction'>]

    """
    folder_name = str(folder_name)
    files = sorted(listdir(folder_name))
    domain_file_name = None
    utility_file_names = []
    for f in files:
        if not f.endswith(".xml") or f.endswith("pareto.xml"):
            continue
        full_name = folder_name + "/" + f
        root = ET.parse(full_name).getroot()

        if root.tag == "negotiation_template":
            domain_file_name = full_name
        elif root.tag == "utility_space":
            utility_file_names.append(full_name)
    return load_genius_domain(
        domain_file_name=domain_file_name,
        utility_file_names=utility_file_names,
        agent_factories=agent_factories,
        force_numeric=force_numeric,
        force_single_issue=force_single_issue,
        cache_and_discretize_outcomes=cache_and_discretize_outcomes,
        max_n_outcomes=max_n_outcomes,
        n_discretization=n_discretization,
        keep_issue_names=keep_issue_names,
        keep_value_names=keep_value_names,
        normalize_utilities=normalize_utilities,
        n_steps=n_steps,
        time_limit=time_limit,
        max_n_agents=max_n_agents,
        dynamic_entry=dynamic_entry,
        safe_parsing=safe_parsing,
        ignore_reserved=ignore_reserved,
        ignore_discount=ignore_discount,
    )


def find_domain_and_utility_files(folder_name) -> Tuple[str, List[str]]:
    """Finds the domain and utility_function files in a folder
    """
    files = sorted(listdir(folder_name))
    domain_file_name = None
    utility_file_names = []
    for f in files:
        if not f.endswith(".xml") or f.endswith("pareto.xml"):
            continue
        full_name = folder_name + "/" + f
        root = ET.parse(full_name).getroot()

        if root.tag == "negotiation_template":
            domain_file_name = full_name
        elif root.tag == "utility_space":
            utility_file_names.append(full_name)
    return domain_file_name, utility_file_names


def convert_genius_domain(
    src_domain_file_name: str,
    dst_domain_file_name: str,
    src_utility_file_names: Optional[List[str]] = None,
    dst_utility_file_names: Optional[List[str]] = None,
    force_single_issue=False,
    cache_and_discretize_outcomes=False,
    max_n_outcomes: int = 1e6,
    n_discretization: Optional[int] = None,
    keep_issue_names=True,
    keep_value_names=True,
    normalize_utilities=True,
    safe_parsing=False,
) -> bool:
    if (
        not force_single_issue
        and not cache_and_discretize_outcomes
        and keep_issue_names
        and keep_value_names
        and not normalize_utilities
    ):
        # no need to do anything, just copy
        shutil.copy(src=src_domain_file_name, dst=dst_domain_file_name)
        for src, dst in zip(src_utility_file_names, dst_utility_file_names):
            shutil.copy(src=src, dst=dst)
        return True
    issues, issues_details, mechanism = None, None, None
    if src_domain_file_name is not None:
        issues_details, _ = Issue.from_genius(
            src_domain_file_name,
            force_single_issue=False,
            keep_issue_names=True,
            keep_value_names=True,
            safe_parsing=safe_parsing,
            n_discretization=n_discretization,
        )
        if force_single_issue:
            issues, _ = Issue.from_genius(
                src_domain_file_name,
                force_single_issue=force_single_issue,
                keep_issue_names=keep_issue_names,
                keep_value_names=keep_value_names,
                max_n_outcomes=max_n_outcomes,
                n_discretization=n_discretization,
            )
        else:
            issues = issues_details
        if issues is None:
            return False
        Issue.to_genius(
            issues=issues, file_name=dst_domain_file_name, enumerate_integer=True
        )

    if src_utility_file_names is None:
        src_utility_file_names = []
    for ufname, dstfname in zip(src_utility_file_names, dst_utility_file_names):
        utility, discount_factor = UtilityFunction.from_genius(
            file_name=ufname,
            force_single_issue=force_single_issue,
            keep_issue_names=keep_issue_names,
            keep_value_names=keep_value_names,
            normalize_utility=normalize_utilities,
            domain_issues=issues_details,
            safe_parsing=safe_parsing,
            max_n_outcomes=max_n_outcomes,
        )
        if utility is None:
            return False
        UtilityFunction.to_genius(
            u=utility,
            discount_factor=discount_factor,
            issues=issues,
            file_name=dstfname,
        )

    return True


def convert_genius_domain_from_folder(
    src_folder_name: str, dst_folder_name: str, **kwargs
) -> bool:
    """Loads a genius domain from a folder. See ``load_genius_domain`` for more details.



    """
    os.makedirs(dst_folder_name, exist_ok=True)
    files = sorted(listdir(src_folder_name))
    domain_file_name = None
    utility_file_names = []
    for f in files:
        if not f.endswith(".xml") or f.endswith("pareto.xml"):
            continue
        full_name = src_folder_name + "/" + f
        root = ET.parse(full_name).getroot()

        if root.tag == "negotiation_template":
            domain_file_name = full_name
        elif root.tag == "utility_space":
            utility_file_names.append(full_name)
    success = convert_genius_domain(
        src_domain_file_name=domain_file_name,
        dst_domain_file_name=os.path.join(
            dst_folder_name, os.path.basename(domain_file_name)
        ),
        src_utility_file_names=utility_file_names,
        dst_utility_file_names=[
            os.path.join(dst_folder_name, os.path.basename(_))
            for _ in utility_file_names
        ],
        **kwargs
    )

    if not success:
        if len(os.listdir(dst_folder_name)) == 0:
            shutil.rmtree(dst_folder_name, ignore_errors=True)

    return success



# -*- coding: utf-8 -*-
"""A framework for conducting multi-strand multilateral asynchronous negotiations on multiple issues."""
__author__ = """Yasser Mohammad"""
__email__ = "yasserfarouk@gmail.com"
__version__ = "0.2.24"

import json
import pathlib


NEGMAS_CONFIG = {}
"""Global configuration parameters for NegMAS"""
CONFIG_KEY_JNEGMAS_JAR = "jnegmas_jar"
"""Key name for the JNegMAS jar in `NEGMAS_CONFIG`"""
CONFIG_KEY_GENIUS_BRIDGE_JAR = "genius_bridge_jar"
"""Key name for the Genius bridge jar in `NEGMAS_CONFIG`"""


# loading config file if any
__conf_path = pathlib.Path("~/negmas/config.json").expanduser().absolute()

if __conf_path.exists():
    with open(__conf_path, "r") as f:
        NEGMAS_CONFIG = json.load(f)

from .common import *
from .outcomes import *
from .utilities import *
from .negotiators import *
from .mechanisms import *
from .sao import *
from .inout import *
from .genius import *
from .situated import *

__all__ = (
    common.__all__
    + outcomes.__all__
    + utilities.__all__
    + negotiators.__all__
    + mechanisms.__all__
    + sao.__all__
    + inout.__all__
    + genius.__all__
    + situated.__all__
    + ["generics", "helpers", "events", "apps", "tournaments"]
)



"""The base class agent needed for all SCML agents."""
import itertools
import math
from abc import abstractmethod
from collections import defaultdict
from typing import Dict, List, Optional, Any, TYPE_CHECKING

from negmas import UtilityFunction
from ...common import AgentMechanismInterface
from ...negotiators import Negotiator
from ...outcomes import Issue
from ...situated import Agent
from ...situated import Contract, Breach
from .common import (
    ManufacturingProfileCompiled,
    ProductManufacturingInfo,
    Process,
    Product,
    Loan,
    CFP,
    FinancialReport,
)

if TYPE_CHECKING:
    from .awi import SCMLAWI

__all__ = ["SCMLAgent"]


class SCMLAgent(Agent):
    """The base for all SCM Agents"""

    def __init__(self, name: str = None):
        super().__init__(name=name)
        self.line_profiles: Dict[int, ManufacturingProfileCompiled] = {}
        """A mapping specifying for each `Line` index, all the profiles used to run it in the factory"""
        self.process_profiles: Dict[int, ManufacturingProfileCompiled] = {}
        """A mapping specifying for each `Process` index, all the profiles used to run it in the factory"""
        self.producing: Dict[int, List[ProductManufacturingInfo]] = defaultdict(list)
        """Mapping from a product to all manufacturing processes that can generate it"""
        self.consuming: Dict[int, List[ProductManufacturingInfo]] = defaultdict(list)
        """Mapping from a product to all manufacturing processes that can consume it"""
        self.compiled_profiles: List[ManufacturingProfileCompiled] = []
        """All the profiles to be used by the factory belonging to this agent compiled to use indices"""
        self.immediate_negotiations = False
        """Whether or not negotiations start immediately upon registration (default is to start on the next production 
        step)"""
        self.negotiation_speed_multiple: int = 1
        """The number of negotiation rounds (steps) conducted in a single production step"""
        self.transportation_delay: int = 0
        """Transportation delay in the system. Default is zero"""
        self.products: List[Product] = []
        """List of products in the system"""
        self.processes: List[Process] = []
        """List of processes in the system"""

    @property
    def awi(self) -> "SCMLAWI":
        """Returns the Agent-World-Interface through which the agent does all of its actions in the world.

        A single excption is request_negotiation for which it is recommended to actually call the helper method
        on the agent itself instead of directly calling the AWI version."""
        return self._awi

    @awi.setter
    def awi(self, awi: "SCMLAWI"):
        """Sets the AWI. Not to be used by agents. Only used by the world simulation itself."""
        self._awi = awi

    def init_(self):
        """The initialization function called by the world directly.

        It does the following actions by default:

            1. copies some of the static world settings to the agent to make them available without calling the AWI.
            2. prepares production related properties like producing, consuming, line_profiles, compiled_profiles, etc.
            3. registers interest in all products that the agent can produce or consume in its factory.
            4. finally it calls any custom initialization logic implemented in `init`()

        See Also:

            `init`, `step`

        """
        # noinspection PyUnresolvedReferences
        self.products = self.awi.products  # type: ignore
        # noinspection PyUnresolvedReferences
        self.processes = self.awi.processes  # type: ignore
        self.negotiation_speed_multiple = self.awi.bb_read(
            "settings", "negotiation_speed_multiple"
        )
        self.immediate_negotiations = self.awi.bb_read(
            "settings", "immediate_negotiations"
        )
        self.transportation_delay = self.awi.bb_read(
            section="settings", key="transportation_delay"
        )

        factory = self.awi.state
        if factory is None:
            raise ValueError("Cannot init any SCMLAgent without specifying a factory")
        profiles = factory.profiles
        self.line_profiles = defaultdict(list)
        self.process_profiles = defaultdict(list)
        self.compiled_profiles = []
        self.producing = defaultdict(list)
        self.consuming = defaultdict(list)
        p2i = dict(zip(self.processes, range(len(self.processes))))
        for index, profile in enumerate(profiles):
            compiled = ManufacturingProfileCompiled.from_manufacturing_profile(
                profile=profile, process2ind=p2i
            )
            self.compiled_profiles.append(compiled)
            self.line_profiles[profile.line].append(compiled)
            self.process_profiles[profile.process].append(compiled)
            process = profile.process
            for outpt in process.outputs:
                step = int(math.ceil(outpt.step * profile.n_steps))
                self.producing[outpt.product].append(
                    ProductManufacturingInfo(
                        profile=index, quantity=outpt.quantity, step=step
                    )
                )

            for inpt in process.inputs:
                step = int(math.floor(inpt.step * profile.n_steps))
                self.consuming[inpt.product].append(
                    ProductManufacturingInfo(
                        profile=index, quantity=inpt.quantity, step=step
                    )
                )
        self.awi.register_interest(
            list(set(itertools.chain(self.producing.keys(), self.consuming.keys())))
        )
        self.init()

    def can_expect_agreement(self, cfp: "CFP", margin: int):
        """
        Checks if it is possible in principle to get an agreement on this CFP by the time it becomes executable

        Args:
            margin:
            cfp:

        Returns:

        """
        return (
            cfp.max_time
            >= self.awi.current_step + 1 - int(self.immediate_negotiations) + margin
        )

    def _create_annotation(self, cfp: "CFP", partner: str = None):
        """Creates full annotation based on a cfp that the agent is receiving

        Args:
            cfp: The call for proposal to create annotation about
            partner: The partner who requested the negotiation

        Remarks:

            - If the annotation is to be created for a CFP that was published by self, partner must be passed

        """
        if self.id == cfp.publisher and partner is None:
            raise ValueError(
                f"{self.id} published {str(cfp)} and create annotation is called without 'partner'"
            )
        if self.id == cfp.publisher:
            partners = [self.id, partner]
            non_publisher = partner
        else:
            partners = [self.id, cfp.publisher]
            non_publisher = self.id

        annotation = {"cfp": cfp, "partners": partners}
        if cfp.is_buy:
            annotation["seller"] = non_publisher
            annotation["buyer"] = cfp.publisher
        else:
            annotation["buyer"] = non_publisher
            annotation["seller"] = cfp.publisher
        return annotation

    def _respond_to_negotiation_request(
        self,
        initiator: str,
        partners: List[str],
        issues: List[Issue],
        annotation: Dict[str, Any],
        mechanism: AgentMechanismInterface,
        role: Optional[str],
        req_id: Optional[str],
    ) -> Optional[Negotiator]:
        """
        Called by the mechanism to ask for joining a negotiation. The agent can refuse by returning a None

        Args:
            initiator: The ID of the agent that initiated the negotiation request
            partners: The partner list (will include this agent)
            issues: The list of issues
            annotation: Any annotation specific to this negotiation.
            mechanism: The mechanism that started the negotiation
            role: The role of this agent in the negotiation
            req_id: The req_id passed to the AWI when starting the negotiation (only to the initiator).

        Returns:
            None to refuse the negotiation or a `Negotiator` object appropriate to the given mechanism to accept it.

        Remarks:

            - It is expected that world designers will introduce a better way to respond and override this function to
              call it

        """
        cfp = annotation["cfp"]
        return self.respond_to_negotiation_request(
            cfp=cfp, partner=initiator if self.id != initiator else cfp.publisher
        )

    def request_negotiation(
        self, cfp: CFP, negotiator: Negotiator = None, ufun: UtilityFunction = None
    ) -> bool:
        """
        Requests a negotiation from the AWI while keeping track of available negotiation requests

        Args:

            cfp:
            negotiator:
            ufun:

        Returns:

            Whether the negotiation request was successful indicating that the partner accepted the negotiation
        """
        # if cfp.publisher == self.id:
        #    return False
        if negotiator is not None and ufun is not None:
            negotiator.utility_function = ufun
        req_id = self._add_negotiation_request_info(
            issues=cfp.issues,
            partners=[self.id, cfp.publisher],
            annotation=None,
            negotiator=negotiator,
            extra=None,
        )
        return self.awi.request_negotiation(cfp=cfp, req_id=req_id)

    # ------------------------------------------------------------------
    # EVENT CALLBACKS (Called by the `World` when certain events happen)
    # ------------------------------------------------------------------

    @abstractmethod
    def on_contract_executed(self, contract: Contract) -> None:
        pass

    @abstractmethod
    def on_contract_breached(
        self, contract: Contract, breaches: List[Breach], resolution: Optional[Contract]
    ) -> None:
        pass

    @abstractmethod
    def confirm_loan(self, loan: Loan, bankrupt_if_rejected: bool) -> bool:
        """called by the world manager to confirm a loan if needed by the buyer of a contract that is about to be
        breached"""

    @abstractmethod
    def on_contract_nullified(
        self, contract: Contract, bankrupt_partner: str, compensation: float
    ) -> None:
        """Will be called whenever a contract the agent is involved in is nullified because another partner went
        bankrupt"""

    @abstractmethod
    def on_agent_bankrupt(self, agent_id: str) -> None:
        """
        Will be called whenever any agent goes bankrupt

        Args:

            agent_id: The ID of the agent that went bankrupt

        Remarks:

            - Agents can go bankrupt in two cases:

                1. Failing to pay one installments of a loan they bought and refusing (or being unable to) get another
                   loan to pay it.
                2. Failing to pay a penalty on a sell contract they failed to honor (and refusing or being unable to get
                   a loan to pay for it).

            - All built-in agents ignore this call and they use the bankruptcy list ONLY to decide whether or not to
              negotiate in their `on_new_cfp` and `respond_to_negotiation_request` callbacks by pulling the
              bulletin-board using the helper function `is_bankrupt` of their AWI.
        """

    @abstractmethod
    def confirm_partial_execution(
        self, contract: Contract, breaches: List[Breach]
    ) -> bool:
        """Will be called whenever a contract cannot be fully executed due to breaches by the other partner.

        Args:

            contract: The contract that was breached
            breaches: A list of all the breaches committed.

        Remarks:

            - Will not be called if both partners committed breaches.
        """

    @abstractmethod
    def confirm_contract_execution(self, contract: Contract) -> bool:
        """Called before executing any agreement"""
        return True

    @abstractmethod
    def respond_to_negotiation_request(
        self, cfp: "CFP", partner: str
    ) -> Optional[Negotiator]:
        """Called when a prospective partner requests a negotiation to start"""

    @abstractmethod
    def on_new_cfp(self, cfp: "CFP"):
        """Called when a new CFP for a product for which the agent registered interest is published"""

    @abstractmethod
    def on_remove_cfp(self, cfp: "CFP"):
        """Called when a new CFP for a product for which the agent registered interest is removed"""

    @abstractmethod
    def on_new_report(self, report: FinancialReport):
        """Called whenever a financial report is published.

        Args:

            report: The financial report giving details of the standing of an agent at some time (see `FinancialReport`)

        Remarks:

            - Agents must opt-in to receive these calls by calling `receive_financial_reports` on their AWI
        """

    @abstractmethod
    def on_inventory_change(self, product: int, quantity: int, cause: str) -> None:
        """
        Received whenever something moves in or out of the factory's storage

        Args:
            product: Product index.
            quantity: Negative value for products moving out and positive value for products moving in
            cause: The cause of the change. Possibilities include:

                   - contract: Contract execution
                   - insurance: Received from insurance company
                   - bankruptcy: Liquidated due to bankruptcy
                   - transport: Arrival of goods (when transportation delay in the system is > 0).
        """

    @abstractmethod
    def on_cash_transfer(self, amount: float, cause: str) -> None:
        """
        Received whenever money is transferred to the factory or from it.

        Args:
            amount: Amount of money (negative for transfers out of the factory, positive for transfers to it).
            cause: The cause of the change. Possibilities include:

                   - contract: Contract execution
                   - insurance: Received from insurance company
                   - bankruptcy: Liquidated due to bankruptcy
                   - transfer: Arrival of transferred money (when transfer delay in the system is > 0).
        """



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
Autogenerated by poetry-setup:
https://github.com/orsinium/poetry-setup
"""
# IMPORTANT: this file is autogenerated. Do not edit it manually.
# All changes will be lost after `poetry-setup` command execution.
# ----------------------------------------------------------------
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path
# io.open is needed for projects that support Python 2.7
# It ensures open() defaults to text mode with universal newlines,
# and accepts an argument to specify the text encoding
# Python 3 only projects can skip this import
from io import open
here = path.abspath(path.dirname(__file__))
# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()
with open(path.join(here, 'HISTORY.rst'), encoding='utf-8') as f:
    long_description += f.read()
# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.
setup(
    # https://packaging.python.org/specifications/core-metadata/#name
    name='negmas',  # Required
    # https://www.python.org/dev/peps/pep-0440/
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.2.24',  # Required
    # https://packaging.python.org/specifications/core-metadata/#summary
    description="NEGotiations Managed by Agent Simulations",  # Required
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional
    # https://packaging.python.org/specifications/core-metadata/#description-content-type-optional
    long_description_content_type='text/markdown',  # Optional (see note above)
    url="https://github.com/yasserfarouk/negmas",  # Optional
    author="Yasser Mohammad",  # Optional
    author_email="yasserfarouk@gmail.com",  # Optional
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],  # Optional
    keywords=' '.join(['negotiation', 'mas', 'multi-agent', 'simulation',
                       'AI']),  # Optional
    packages=find_packages(exclude=[
        'negmas/**/*.pyc',
        'nemgas/**/__pycache__',
        'config',
        'docs',
        'etc',
        'notebooks',
        'tests',
    ]),  # Required
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'Click (>=6.0)',
        'dataclasses; python_version < "3.7"',
        'PyYAML (==5.1b1)',
        'progressbar2 (>=3.39)',
        'typing_extensions (>=3.7)',
        'pytest-runner (>=4.4)',
        'pandas (>=0.24.1)',
        'scipy (>=1.2)',
        'numpy (>=1.16)',
        'stringcase',
        'py4j',
        'colorlog',
        'inflect',
        'matplotlib',
        'setuptools (>=40.8.0)',
        'tabulate',
        'typing',
        'tqdm (>=4.31.1)',
        'click-config-file (>=0.4.5,<0.5.0)',
    ],  # Optional
    # https://setuptools.readthedocs.io/en/latest/setuptools.html#dependencies-that-aren-t-in-pypi
    dependency_links=[],  # Optional
    # https://stackoverflow.com/a/16576850
    include_package_data=True,
    entry_points={  # Optional
        'console_scripts': [
            'negmas=negmas.scripts.app:cli',
        ],
    },
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    project_urls={  # Optional
        'homepage': 'https://github.com/yasserfarouk/negmas',
    },
    python_requires='>=3.6',
)



#!/usr/bin/env python
"""The NegMAS universal command line tool"""
import json
import os
import traceback
import urllib.request
import warnings
from functools import partial
from math import factorial
from pathlib import Path
from pprint import pformat
from time import perf_counter
import numpy as np

import click
import click_config_file
import pandas as pd
import progressbar
import yaml
from tabulate import tabulate

import negmas
from negmas import save_stats
from negmas.apps.scml import *
from negmas.apps.scml.utils import anac2019_sabotage
from negmas.helpers import humanize_time, unique_name, camel_case
from negmas.java import init_jnegmas_bridge, jnegmas_bridge_is_running

try:
    # disable a warning in yaml 1b1 version
    yaml.warnings({"YAMLLoadWarning": False})
except:
    pass

n_completed = 0
n_total = 0

JNEGMAS_JAR_NAME = "jnegmas-0.2.6-all.jar"
GENIUS_JAR_NAME = "genius-8.0.4-bridge.jar"


def print_progress(_, i, n) -> None:
    """Prints the progress of a tournament"""
    global n_completed, n_total
    n_completed = i + 1
    n_total = n
    print(
        f"{n_completed:04} of {n:04} worlds completed ({n_completed / n:0.2%})",
        flush=True,
    )


def print_world_progress(world) -> None:
    """Prints the progress of a world"""
    step = world.current_step + 1
    s = (
        f"World# {n_completed:04}: {step:04}  of {world.n_steps:04} "
        f"steps completed ({step / world.n_steps:0.2f}) "
    )
    if n_total > 0:
        s += f"TOTAL: ({n_completed + step / world.n_steps / n_total:0.2f})"
    print(s, flush=True)


click.option = partial(click.option, show_default=True)


@click.group()
def cli():
    pass


@cli.command(help="Run a tournament between candidate agent types")
@click.option(
    "--name",
    "-n",
    default="random",
    help='The name of the tournament. The special value "random" will result in a random name',
)
@click.option("--steps", "-s", default=100, help="Number of steps.")
@click.option(
    "--ttype",
    "--tournament-type",
    "--tournament",
    default="anac2019collusion",
    type=click.Choice(["anac2019collusion", "anac2019std", "anac2019sabotage"]),
    help="The config to use. Default is ANAC 2019. Options supported are anac2019std, anac2019collusion, "
    "anac2019sabotage",
)
@click.option(
    "--timeout",
    "-t",
    default=0,
    type=int,
    help="Timeout after the given number of seconds (0 for infinite)",
)
@click.option(
    "--configs",
    default=5,
    type=int,
    help="Number of unique configurations to generate.",
)
@click.option("--runs", default=2, help="Number of runs for each configuration")
@click.option(
    "--max-runs",
    default=-1,
    type=int,
    help="Maximum total number of runs. Zero or negative numbers mean no limit",
)
@click.option(
    "--agents",
    default=5,
    type=int,
    help="Number of agents per competitor (not used for anac2019std in which this is preset to 1).",
)
@click.option(
    "--factories",
    default=5,
    type=int,
    help="Minimum numbers of factories to have per level.",
)
@click.option(
    "--competitors",
    default="DoNothingFactoryManager;GreedyFactoryManager",
    help="A semicolon (;) separated list of agent types to use for the competition.",
)
@click.option(
    "--jcompetitors",
    "--java-competitors",
    default="",
    help="A semicolon (;) separated list of agent types to use for the competition.",
)
@click.option(
    "--non-competitors",
    default="GreedyFactoryManager",
    help="A semicolon (;) separated list of agent types to exist in the worlds as non-competitors "
    "(their scores will not be calculated).",
)
@click.option(
    "--parallel/--serial",
    default=True,
    help="Run a parallel/serial tournament on a single machine",
)
@click.option(
    "--distributed/--single-machine",
    default=False,
    help="Run a distributed tournament using dask",
)
@click.option(
    "--log",
    "-l",
    type=click.Path(dir_okay=True, file_okay=False),
    default="~/negmas/logs/tournaments",
    help="Default location to save logs (A folder will be created under it)",
)
@click.option(
    "--verbosity",
    default=1,
    type=int,
    help="verbosity level (from 0 == silent to 1 == world progress)",
)
@click.option("--configs-only/--run", default=False, help="configs_only")
@click.option(
    "--reveal-names/--hidden-names",
    default=True,
    help="Reveal agent names (should be used only for " "debugging)",
)
@click.option(
    "--ip",
    default="127.0.0.1",
    help="The IP address for a dask scheduler to run the distributed tournament."
    " Effective only if --distributed",
)
@click.option(
    "--port",
    default=8786,
    type=int,
    help="The IP port number a dask scheduler to run the distributed tournament."
    " Effective only if --distributed",
)
@click.option(
    "--log-ufuns/--no-ufun-logs",
    default=False,
    help="Log ufuns into their own CSV file. Only effective if --debug is given",
)
@click.option(
    "--log-negs/--no-neg-logs",
    default=False,
    help="Log all negotiations. Only effective if --debug is given",
)
@click.option(
    "--compact/--debug",
    default=True,
    help="If True, effort is exerted to reduce the memory footprint which"
    "includes reducing logs dramatically.",
)
@click_config_file.configuration_option()
def tournament(
    name,
    steps,
    parallel,
    distributed,
    ttype,
    timeout,
    log,
    verbosity,
    configs_only,
    reveal_names,
    ip,
    port,
    runs,
    configs,
    max_runs,
    competitors,
    jcompetitors,
    non_competitors,
    compact,
    factories,
    agents,
    log_ufuns,
    log_negs,
):
    if timeout <= 0:
        timeout = None
    if name == "random":
        name = None
    if max_runs <= 0:
        max_runs = None
    if compact:
        log_ufuns = False

    if not compact:
        if not reveal_names:
            print(
                "You are running the tournament with --debug. Will reveal agent types in their names"
            )
        reveal_names = True
        verbosity = max(1, verbosity)

    worlds_per_config = (
        None if max_runs is None else int(round(max_runs / (configs * runs)))
    )

    parallelism = "distributed" if distributed else "parallel" if parallel else "serial"

    non_competitors = non_competitors.split(";")
    for i, cp in enumerate(non_competitors):
        if "." not in cp:
            non_competitors[i] = "negmas.apps.scml.factory_managers." + cp

    all_competitors = competitors.split(";")
    for i, cp in enumerate(all_competitors):
        if "." not in cp:
            all_competitors[i] = "negmas.apps.scml.factory_managers." + cp
    all_competitors_params = [dict() for _ in range(len(all_competitors))]
    if jcompetitors is not None and len(jcompetitors) > 0:
        jcompetitor_params = [{"java_class_name": _} for _ in jcompetitors.split(";")]
        for jp in jcompetitor_params:
            if "." not in jp["java_class_name"]:
                jp["java_class_name"] = (
                    "jnegmas.apps.scml.factory_managers." + jp["java_class_name"]
                )
        jcompetitors = ["negmas.apps.scml.JavaFactoryManager"] * len(jcompetitor_params)
        all_competitors += jcompetitors
        all_competitors_params += jcompetitor_params
        print("You are using some Java agents. The tournament MUST run serially")
        parallelism = "serial"
        if not jnegmas_bridge_is_running():
            print(
                "Error: You are using java competitors but jnegmas bridge is not running\n\nTo correct this issue"
                " run the following command IN A DIFFERENT TERMINAL because it will block:\n\n"
                "$ negmas jnegmas"
            )
            exit(1)
    prog_callback = print_world_progress if verbosity > 1 and not distributed else None

    recommended = (
        runs * configs * factorial(len(all_competitors) * (1 if "std" in ttype else 3))
    )
    if worlds_per_config is not None and worlds_per_config < 1:
        print(
            f"You need at least {(configs * runs)} runs even with a single permutation of managers."
            f".\n\nSet --max-runs to at least {(configs * runs)} (Recommended {recommended})"
        )
        return

    if max_runs is not None and max_runs < recommended:
        print(
            f"You are running {max_runs} worlds only but it is recommended to set {max_runs} to at least "
            f"{recommended}. Will continue"
        )

    if ttype == "anac2019std":
        agents = 1

    if worlds_per_config is None:
        n_agents_per_competitor = agents
        n_comp = len(all_competitors) if ttype != "anac2019sabotage" else 2
        n_worlds = factorial(n_agents_per_competitor * n_comp) * runs * configs
        if n_worlds > 100:
            print(
                f"You are running the maximum possible number of permutations for each configuration. This is roughly"
                f" {n_worlds} simulations (each for {steps} steps). That will take a VERY long time."
                f"\n\nYou can reduce the number of simulations by setting --configs (currently {configs}) or --runs"
                f" (currently {runs}) to a lower value. If you are running a collusion competition, you can reduce"
                f" the number of simulations "
                f"{factorial(n_agents_per_competitor * n_comp)/factorial(n_comp)} times "
                f"by running a standard competition (--ttype=anac2019std)."
                f"\nFinally, you can limit the maximum number of worlds to run by setting --max-runs=integer."
            )
            if ttype != "anac2019std":
                n_new = factorial(n_comp) * runs * configs
                print(
                    f"If you use --ttype=anac2019std (standard competition), the number of simulations will be {n_new}"
                )
            if (
                not input(f"Are you sure you want to run {n_worlds} simulations?")
                .lower()
                .startswith("y")
            ):
                exit(0)

    if len(jcompetitors) > 0:
        print("You are using java-competitors. The tournament will be run serially")
        parallelism = "serial"

    start = perf_counter()
    if ttype.lower() == "anac2019std":
        results = anac2019_std(
            competitors=all_competitors,
            competitor_params=all_competitors_params,
            agent_names_reveal_type=reveal_names,
            tournament_path=log,
            total_timeout=timeout,
            parallelism=parallelism,
            scheduler_ip=ip,
            scheduler_port=port,
            world_progress_callback=prog_callback,
            name=name,
            verbose=verbosity > 0,
            n_runs_per_world=runs,
            n_configs=configs,
            max_worlds_per_config=worlds_per_config,
            non_competitors=non_competitors,
            configs_only=configs_only,
            min_factories_per_level=factories,
            n_steps=steps,
            compact=compact,
            log_ufuns=log_ufuns,
            log_negotiations=log_negs,
        )
    elif ttype.lower() in ("anac2019collusion", "anac2019"):
        results = anac2019_collusion(
            competitors=all_competitors,
            competitor_params=all_competitors_params,
            agent_names_reveal_type=reveal_names,
            n_agents_per_competitor=agents,
            tournament_path=log,
            total_timeout=timeout,
            parallelism=parallelism,
            scheduler_ip=ip,
            scheduler_port=port,
            world_progress_callback=prog_callback,
            name=name,
            verbose=verbosity > 0,
            n_runs_per_world=runs,
            n_configs=configs,
            max_worlds_per_config=worlds_per_config,
            non_competitors=non_competitors,
            configs_only=configs_only,
            min_factories_per_level=factories,
            n_steps=steps,
            compact=compact,
            log_ufuns=log_ufuns,
            log_negotiations=log_negs,
        )
    else:
        results = anac2019_sabotage(
            competitors=all_competitors,
            competitor_params=all_competitors_params,
            agent_names_reveal_type=reveal_names,
            n_agents_per_competitor=agents,
            tournament_path=log,
            total_timeout=timeout,
            parallelism=parallelism,
            scheduler_ip=ip,
            scheduler_port=port,
            world_progress_callback=prog_callback,
            name=name,
            verbose=verbosity > 0,
            n_runs_per_world=runs,
            n_configs=configs,
            max_worlds_per_config=worlds_per_config,
            non_competitors=non_competitors,
            configs_only=configs_only,
            min_factories_per_level=factories,
            n_steps=steps,
            compact=compact,
            log_ufuns=log_ufuns,
            log_negotiations=log_negs,
        )
    if configs_only:
        print(f"Saved all configs to {str(results)}")
        print(f"Finished in {humanize_time(perf_counter() - start)} [config-only]")
        return
    print(tabulate(results.total_scores, headers="keys", tablefmt="psql"))
    print(f"Finished in {humanize_time(perf_counter() - start)} [{parallelism}]")


@cli.command(help="Run an SCML world simulation")
@click.option("--steps", default=100, type=int, help="Number of steps.")
@click.option(
    "--levels",
    default=3,
    type=int,
    help="Number of intermediate production levels (processes). "
    "-1 means a single product and no factories.",
)
@click.option("--neg-speedup", default=21, help="Negotiation Speedup.")
@click.option(
    "--negotiator",
    default=DEFAULT_NEGOTIATOR,
    help="Negotiator type to use for builtin agents.",
)
@click.option(
    "--min-consumption",
    default=3,
    type=int,
    help="The minimum number of units consumed by each consumer at every " "time-step.",
)
@click.option(
    "--max-consumption",
    default=5,
    type=int,
    help="The maximum number of units consumed by each consumer at every " "time-step.",
)
@click.option(
    "--agents",
    default=5,
    type=int,
    help="Number of agents (miners/negmas.consumers) per production level",
)
@click.option("--horizon", default=20, type=int, help="Consumption horizon.")
@click.option("--transport", default=0, type=int, help="Transportation Delay.")
@click.option("--time", default=60 * 90, type=int, help="Total time limit.")
@click.option(
    "--neg-time", default=60 * 4, type=int, help="Time limit per single negotiation"
)
@click.option(
    "--neg-steps", default=20, type=int, help="Number of rounds per single negotiation"
)
@click.option(
    "--sign",
    default=1,
    type=int,
    help="The default delay between contract conclusion and signing",
)
@click.option(
    "--guaranteed",
    default=False,
    help="Whether to only sign contracts that are guaranteed not to cause " "breaches",
)
@click.option("--lines", default=10, help="The number of lines per factory")
@click.option(
    "--retrials",
    default=5,
    type=int,
    help="The number of times an agent re-tries on failed negotiations",
)
@click.option(
    "--use-consumer/--no-consumer",
    default=True,
    help="Use internal consumer object in factory managers",
)
@click.option(
    "--max-insurance",
    default="inf",
    type=float,
    help="Use insurance against partner in factory managers up to this premium. Pass zero for never buying insurance"
    " and a 'inf' (without quotes) for infinity.",
)
@click.option(
    "--riskiness", default=0.0, help="How risky is the default factory manager"
)
@click.option(
    "--competitors",
    default="GreedyFactoryManager",
    help="A semicolon (;) separated list of agent types to use for the competition.",
)
@click.option(
    "--jcompetitors",
    "--java-competitors",
    default="",
    help="A semicolon (;) separated list of agent types to use for the competition.",
)
@click.option(
    "--log",
    type=click.Path(file_okay=False, dir_okay=True),
    default="~/negmas/logs",
    help="Default location to save logs (A folder will be created under it)",
)
@click.option(
    "--log-ufuns/--no-ufun-logs",
    default=False,
    help="Log ufuns into their own CSV file. Only effective if --debug is given",
)
@click.option(
    "--log-negs/--no-neg-logs",
    default=False,
    help="Log all negotiations. Only effective if --debug is given",
)
@click.option(
    "--compact/--debug",
    default=False,
    help="If True, effort is exerted to reduce the memory footprint which"
    "includes reducing logs dramatically.",
)
@click_config_file.configuration_option()
def scml(
    steps,
    levels,
    neg_speedup,
    negotiator,
    agents,
    horizon,
    min_consumption,
    max_consumption,
    transport,
    time,
    neg_time,
    neg_steps,
    sign,
    guaranteed,
    lines,
    retrials,
    use_consumer,
    max_insurance,
    riskiness,
    competitors,
    jcompetitors,
    log,
    compact,
    log_ufuns,
    log_negs,
):
    if max_insurance < 0:
        warnings.warn(
            f"Negative max insurance ({max_insurance}) is deprecated. Set --max-insurance=inf for always "
            f"buying and --max-insurance=0.0 for never buying. Will continue assuming --max-insurance=inf"
        )
        max_insurance = float("inf")

    if "." not in negotiator:
        negotiator = "negmas.sao." + negotiator

    params = {
        "steps": steps,
        "levels": levels,
        "neg_speedup": neg_speedup,
        "negotiator": negotiator,
        "agents": agents,
        "horizon": horizon,
        "min_consumption": min_consumption,
        "max_consumption": max_consumption,
        "transport": transport,
        "time": time,
        "neg_time": neg_time,
        "neg_steps": neg_steps,
        "sign": sign,
        "guaranteed": guaranteed,
        "lines": lines,
        "retrials": retrials,
        "use_consumer": use_consumer,
        "max_insurance": max_insurance,
        "riskiness": riskiness,
    }
    if compact:
        log_ufuns = False
        log_negs = False
    neg_speedup = neg_speedup if neg_speedup is not None and neg_speedup > 0 else None
    if min_consumption == max_consumption:
        consumption = min_consumption
    else:
        consumption = (min_consumption, max_consumption)
    customer_kwargs = {"negotiator_type": negotiator, "consumption_horizon": horizon}
    miner_kwargs = {"negotiator_type": negotiator, "n_retrials": retrials}
    factory_kwargs = {
        "negotiator_type": negotiator,
        "n_retrials": retrials,
        "sign_only_guaranteed_contracts": guaranteed,
        "use_consumer": use_consumer,
        "riskiness": riskiness,
        "max_insurance_premium": max_insurance,
    }
    if log.startswith("~/"):
        log_dir = Path.home() / log[2:]
    else:
        log_dir = Path(log)
    world_name = unique_name(base="scml", add_time=True, rand_digits=0)
    log_dir = log_dir / world_name
    log_dir = log_dir.absolute()
    os.makedirs(log_dir, exist_ok=True)
    log_file_name = str(log_dir / "log.txt")

    exception = None

    def _no_default(s):
        return s.startswith("negmas.apps.scml") and s.endswith("GreedyFactoryManager")

    all_competitors = competitors.split(";")
    for i, cp in enumerate(all_competitors):
        if "." not in cp:
            all_competitors[i] = "negmas.apps.scml.factory_managers." + cp
    all_competitors_params = [
        dict() if _no_default(_) else factory_kwargs for _ in all_competitors
    ]
    if jcompetitors is not None and len(jcompetitors) > 0:
        jcompetitor_params = [{"java_class_name": _} for _ in jcompetitors.split(";")]
        for jp in jcompetitor_params:
            if "." not in jp["java_class_name"]:
                jp["java_class_name"] = (
                    "jnegmas.apps.scml.factory_managers." + jp["java_class_name"]
                )
        jcompetitors = ["negmas.apps.scml.JavaFactoryManager"] * len(jcompetitor_params)
        all_competitors += jcompetitors
        all_competitors_params += jcompetitor_params
        print("You are using some Java agents. The tournament MUST run serially")
        parallelism = "serial"
        if not jnegmas_bridge_is_running():
            print(
                "Error: You are using java competitors but jnegmas bridge is not running\n\nTo correct this issue"
                " run the following command IN A DIFFERENT TERMINAL because it will block:\n\n"
                "$ negmas jnegmas"
            )
            exit(1)

    world = SCMLWorld.chain_world(
        log_file_name=log_file_name,
        n_steps=steps,
        negotiation_speed=neg_speedup,
        n_intermediate_levels=levels,
        n_miners=agents,
        n_consumers=agents,
        n_factories_per_level=agents,
        consumption=consumption,
        consumer_kwargs=customer_kwargs,
        miner_kwargs=miner_kwargs,
        default_manager_params=factory_kwargs,
        transportation_delay=transport,
        time_limit=time,
        neg_time_limit=neg_time,
        neg_n_steps=neg_steps,
        default_signing_delay=sign,
        n_lines_per_factory=lines,
        compact=compact,
        agent_names_reveal_type=True,
        log_ufuns=log_ufuns,
        manager_types=all_competitors,
        manager_params=all_competitors_params,
        log_negotiations=log_negs,
        log_folder=log_dir,
        name=world_name,
    )
    failed = False
    strt = perf_counter()
    try:
        for i in progressbar.progressbar(range(world.n_steps), max_value=world.n_steps):
            elapsed = perf_counter() - strt
            if world.time_limit is not None and elapsed >= world.time_limit:
                break
            if not world.step():
                break
    except Exception:
        exception = traceback.format_exc()
        failed = True
    elapsed = perf_counter() - strt

    def print_and_log(s):
        world.logdebug(s)
        print(s)

    world.logdebug(f"{pformat(world.stats, compact=True)}")
    world.logdebug(
        f"=================================================\n"
        f"steps: {steps}, horizon: {horizon}, time: {time}, levels: {levels}, agents_per_level: "
        f"{agents}, lines: {lines}, guaranteed: {guaranteed}, negotiator: {negotiator}\n"
        f"consumption: {consumption}"
        f", transport_to: {transport}, sign: {sign}, speedup: {neg_speedup}, neg_steps: {neg_steps}"
        f", retrials: {retrials}"
        f", neg_time: {neg_time}\n"
        f"=================================================="
    )

    save_stats(world=world, log_dir=log_dir, params=params)

    if len(world.saved_contracts) > 0:
        data = pd.DataFrame(world.saved_contracts)
        data = data.sort_values(["delivery_time"])
        data = data.loc[
            data.signed_at >= 0,
            [
                "seller_type",
                "buyer_type",
                "seller_name",
                "buyer_name",
                "delivery_time",
                "unit_price",
                "quantity",
                "product_name",
                "n_neg_steps",
                "signed_at",
            ],
        ]
        data.columns = [
            "seller_type",
            "buyer_type",
            "seller",
            "buyer",
            "t",
            "price",
            "q",
            "product",
            "steps",
            "signed",
        ]
        print_and_log(tabulate(data, headers="keys", tablefmt="psql"))

        try:
            data["product_id"] = np.array([_.id for _ in data["product"].values])
            d2 = (
                data.groupby(["product_id"])
                .apply(
                    lambda x: np.sum(x["price"].values * x["q"].values)
                    / np.sum(x["q"].values)
                )
                .reset_index()
            )
            products = dict(zip([_.id for _ in world.products], world.products))
            d2["Product"] = np.array([products[_] for _ in d2["product_id"]])
            d2 = d2.loc[:, ["Product", 0]]
            d2.columns = ["Product", "Trading Price"]
            print_and_log(tabulate(d2, headers="keys", tablefmt="psql"))
        except:
            pass

        n_executed = sum(world.stats["n_contracts_executed"])
        n_negs = sum(world.stats["n_negotiations"])
        n_contracts = len(world.saved_contracts)
        try:
            agent_scores = sorted(
                [
                    [_.name, world.a2f[_.id].balance]
                    for _ in world.agents.values()
                    if isinstance(_, FactoryManager)
                ],
                key=lambda x: x[1],
                reverse=True,
            )
            agent_scores = pd.DataFrame(
                data=np.array(agent_scores), columns=["Agent", "Final Balance"]
            )
            print_and_log(tabulate(agent_scores, headers="keys", tablefmt="psql"))
        except:
            pass
        winners = [
            f"{_.name} gaining {world.a2f[_.id].balance / world.a2f[_.id].initial_balance - 1.0:0.0%}"
            for _ in world.winners
        ]
        print_and_log(
            f"{n_contracts} contracts :-) [N. Negotiations: {n_negs}, Agreement Rate: "
            f"{world.agreement_rate:0.0%}]"
            f" (rounds/successful negotiation: {world.n_negotiation_rounds_successful:5.2f}, "
            f"rounds/broken negotiation: {world.n_negotiation_rounds_failed:5.2f})"
        )
        print_and_log(
            f"Cancelled: {world.cancellation_rate:0.0%}, Executed: {world.contract_execution_fraction:0.0%}"
            f", Breached: {world.breach_rate:0.0%}, N. Executed: {n_executed}, Business size: "
            f"{world.business_size}\n"
            f"Winners: {winners}\n"
            f"Running Time {humanize_time(elapsed)}"
        )
    else:
        print_and_log("No contracts! :-(")
        print_and_log(f"Running Time {humanize_time(elapsed)}")

    if failed:
        print(exception)
        world.logdebug(exception)
        print(f"FAILED at step {world.current_step} of {world.n_steps}\n")


@cli.command(help="Start the bridge to genius (to use GeniusNegotiator)")
@click.option(
    "--path",
    "-p",
    default="auto",
    help='Path to genius-8.0.4.jar with embedded NegLoader. Use "auto" to '
    "read the path from ~/negmas/config.json"
    "\n\tConfig key is genius_bridge_jar"
    "\nYou can download this jar from: "
    "http://www.yasserm.com/scml/genius-8.0.4-bridge.jar",
)
@click.option(
    "--port",
    "-r",
    default=0,
    help="Port to run the NegLoader on. Pass 0 for the default value",
)
@click.option(
    "--force/--no-force",
    default=False,
    help="Force trial even if an earlier instance exists",
)
def genius(path, port, force):
    negmas.init_genius_bridge(
        path=path if path != "auto" else None, port=port, force=force
    )


@cli.command(help="Start the bridge to JNegMAS (to use Java agents in worlds)")
@click.option(
    "--path",
    "-p",
    default="auto",
    help='Path to jnegmas*.jar with. Use "auto" to '
    "read the path from ~/negmas/config.json."
    "\n\tConfig key is jnegmas_jar"
    "\nYou can download the latest version of this jar from: "
    "http://www.yasserm.com/scml/jnegmas-all.jar",
)
@click.option(
    "--port",
    "-r",
    default=0,
    help="Port to run the jnegmas on. Pass 0 for the default value",
)
@click_config_file.configuration_option()
def jnegmas(path, port):
    init_jnegmas_bridge(path=path if path != "auto" else None, port=port)
    input(
        "Press C^c to quit. You may also need to kill any remaining java processes manually"
    )


def download_and_set(key, url, file_name):
    """
    Downloads a file and sets the corresponding key in ~/negmas/config.json

    Args:
        key: Key name in config.json
        url: URL to download from
        file_name: file name

    Returns:

    """
    config_path = Path.home() / "negmas"
    jar_path = config_path / "files"
    jar_path.mkdir(parents=True, exist_ok=True)
    jar_path = jar_path / file_name
    urllib.request.urlretrieve(url, jar_path)
    config = {}
    config_file = config_path / "config.json"
    if config_file.exists():
        with open(config_file, "r") as f:
            config = json.load(f)
    config[key] = str(jar_path)
    with open(config_file, "w") as f:
        json.dump(config, fp=f, sort_keys=True, indent=4)


@cli.command(help="Downloads jnegmas and updates your settings")
def jnegmas_setup():
    url = f"http://www.yasserm.com/scml/{JNEGMAS_JAR_NAME}"
    print(f"Downloading: {url}", end="", flush=True)
    download_and_set(key="jnegmas_jar", url=url, file_name="jnegmas.jar")
    print(" done successfully")


@cli.command(help="Downloads the genius bridge and updates your settings")
def genius_setup():
    url = f"http://www.yasserm.com/scml/{GENIUS_JAR_NAME}"
    print(f"Downloading: {url}", end="", flush=True)
    download_and_set(key="genius_bridge_jar", url=url, file_name="jnegmas.jar")
    print(" done successfully")


@cli.command(help="Prints NegMAS version")
def version():
    print(negmas.__version__)


if __name__ == "__main__":
    cli()
